""" Bid for Game
    Responder's Bid class
"""
import inspect
from .bidding import Bid, Pass, Double, HandSuit
from .hand import Hand
from bridgeobjects import Suit


class RespondersBid(Hand):
    """Responder's bid class."""

    def __init__(self, hand_cards, board):
        super(RespondersBid, self).__init__(hand_cards, board)
        self.trace = 0

    def suggested_bid(self):
        """Direct control to relevant method and return a Bid object."""
        if self.overcaller_bid_one.level >= 4:
            bid = Pass('0036')
        elif self.opener_bid_one.is_nt:
            bid = self._respond_to_nt()
        elif self.opener_bid_one.level == 1:
            bid = self._respond_to_one_suit()
        elif self.opener_bid_one.name == '2C':
            bid = self._respond_to_two_clubs()
        elif self.opener_bid_one.level == 2:
            bid = self._respond_to_weak_two()
        elif self.opener_bid_one.level == 3:
            bid = self._respond_to_weak_three()
        else:
            bid = Pass('0037')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_nt(self):
        """Return responses to 1NT or 2NT."""
        if self.opener_bid_one.level == 1:
            bid = self._respond_to_one_nt()
        elif self.opener_bid_one.level == 2:
            bid = self._respond_to_two_nt()
        else:
            bid = Pass('0000')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_one_nt(self):
        """Respond to 1NT."""
        if self.overcaller_bid_one.is_value_call:
            bid = self._one_nt_with_overcall()
        elif self.hcp <= 10:
            bid = self._weak_take_out()
        else:
            bid = self._one_nt_test_invite()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _one_nt_with_overcall(self):
        """Return bid after 1NT and overcall."""
        if self._six_card_major_thirteen_points():
            bid = self.suit_bid(4, self.longest_suit, '0185')
        elif self._can_bid_after_one_nt_with_overcall():
            bid = self._trial_bid_suit_after_one_nt_with_overcall()
        elif self._is_semi_balanced_eleven_plus():
            bid = self._try_nt_after_one_nt_with_overcall()
        elif self._is_balanced_game_going() and self.five_card_major:
            bid = self.suit_bid(3, self.longest_suit, '0590')
        elif self._is_balanced_game_going():
            bid = self.nt_bid(3, '0196')
        elif self._is_game_going():
            bid = self.next_level_bid(self.longest_suit, '0128')
        elif self.shape[0] >= 6 and self.hcp >= 10 and self.longest_suit not in self.opponents_suits:
            bid = self.next_level_bid(self.longest_suit, '0791')
        else:
            bid = Pass('0038')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _try_nt_after_one_nt_with_overcall(self):
        """Try a NT after 1NT and overcall."""
        if self.hcp >= 13:
            level = 3
        else:
            level = 2
        if level >= self.nt_level:
            bid = self.nt_bid(level, '0041')
        else:
            bid = Pass('0042')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _weak_take_out(self):
        """Return bid for weak-takeout."""
        if self.shape[0] >= 5:
            bid = self._weak_take_out_bid()
        else:
            bid = Pass('0043')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _weak_take_out_bid(self):
        """Return bid for weak-takeout."""
        suit = self.long_suit(5)
        if self.overcaller_bid_one.is_double:
            bid = self.suit_bid(2, suit, '0044')
        elif (self.shape[0] >= 6 and self.shape[1] >= 5 and self.hcp >= 10
              and self.longest_suit.is_major):
            bid = self.suit_bid(4, self.longest_suit, '0649')
        elif suit != self.club_suit:
            bid = self.suit_bid(2, suit, '0045')
        else:
            bid = Pass('0046')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _one_nt_test_invite(self):
        """Bid with four card major."""
        if 11 <= self.hcp <= 12:
            bid = self._one_nt_test_invite_weak()
        elif self.hcp >= 13:
            bid = self._respond_to_one_nt_strong()
        else:
            bid = Pass('0047')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _one_nt_test_invite_weak(self):
        if self.shape[0] >= 6:
            bid = self._respond_to_one_nt_strong()
        elif self.four_card_major:
            bid = self._one_nt_weak_stayman()
        elif self.nt_level <= 2:
            bid = self.nt_bid(2, '0048')
        else:
            bid = Pass('0049')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _one_nt_weak_stayman(self):
        if self.overcall_made:
            bid = self._one_nt_weak_overcall()
        else:
            bid = self.club_bid(2, '0050')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _one_nt_weak_overcall(self):
        """Return bid after overcall."""
        if self.overcaller_bid_one.is_double and self.hcp >= 10:
            bid = Pass('0051')
        elif self.stoppers_in_bid_suits and self.nt_level <= 2:
            bid = self.nt_bid(2, '0052')
        else:
            bid = Pass('0053')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_one_nt_strong(self):
        """Bid after 1NT strong."""
        if self._four_card_major_after_one_nt():
            bid = self._four_card_major_after_one_nt()
        elif self.shape[0] >= 6:
            bid = self.suit_bid(3, self.longest_suit, '0309')
        else:
            bid = self._strong_nt()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _four_card_major_after_one_nt(self):
        """Bid after 1NT, four card major."""
        if self.spades >= 6:
            bid = self.spade_bid(4, '0054')
        elif self.hearts >= 6:
            bid = self.heart_bid(4, '0055')
        elif self.spades >= 5:
            bid = self.spade_bid(3, '0056')
        elif self.hearts >= 5:
            bid = self.heart_bid(3, '0057')
        elif (not self.overcall_made and
              self.hcp <= 18 and
              self.four_card_major):
            bid = self.club_bid(2, '0058')  # Stayman
        else:
            bid = None
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _strong_nt(self):
        """Bid after 1NT, no four card major."""
        if self.hcp <= 18:
            bid = self.nt_bid(3, '0059')
        elif 19 <= self.hcp <= 20:
            bid = self.nt_bid(4, '0060')
        elif 21 <= self.hcp <= 22:
            bid = self.nt_bid(6, '0061')
        elif 23 <= self.hcp <= 24:
            bid = self.nt_bid(5, '0062')
        else:  # elif self.hcp >= 25:
            bid = self.nt_bid(7, '0063')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_two_nt(self):
        """Responses to 2NT."""
        if self.hcp <= 3:
            bid = Pass('0064')
        elif (self.spades >= 5 and
              self.hearts >= 5 and
              self.shape[0] == self.shape[1] and
              self.bid_history[-1] == 'P'):
            bid = self.next_level_bid(self.spade_suit, '0682')
        elif (self.spades >= 4 and
              self.hearts >= 4 and
              self.shape[1] <= 4 and
              self.bid_history[-1] == 'P' and
              self.next_level(self.club_suit) <= 3):
            bid = self.club_bid(3, '0872')
        elif self.spades >= 6 and self.next_level(self.spade_suit) <= 4:
            bid = self.spade_bid(4, '0068')
        elif self.hearts >= 6 and self.next_level(self.heart_suit) <= 4:
            bid = self.heart_bid(4, '0068')
        elif self.spades >= 5 and self.next_level(self.spade_suit) <= 3:
            bid = self.spade_bid(3, '0068')
        elif self.hearts >= 5 and self.next_level(self.heart_suit) <= 3:
            bid = self.heart_bid(3, '0068')
        elif ((self.spades >= 4 or self.hearts >= 4) and
              self.next_level(self.club_suit) <= 3):
            bid = self.club_bid(3, '0069')
        elif 11 <= self.hcp <= 12 and self.nt_level <= 4:
            bid = self.nt_bid(4, '0070')
        elif 13 <= self.hcp <= 14 and self.nt_level <= 6:
            bid = self.nt_bid(6, '0071')
        elif 15 <= self.hcp <= 16 and self.nt_level <= 5:
            bid = self.nt_bid(5, '0072')
        elif self.hcp >= 17 and self.nt_level <= 7:
            bid = self.nt_bid(7, '0073')
        elif self.nt_level <= 3:  #  self.hcp >= 4:
            bid = self.nt_bid(3, '0074')
        else:
            bid = Double('0000')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_one_suit(self):
        """Respond to one of suit."""
        if self.opener_bid_one.denomination.is_minor:
            bid = self._minor_opening_check_major()
        else:
            bid = self._minor_opening_check_support()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _minor_opening_check_major(self):
        """Respond after one of minor with four card major."""
        if self.four_card_major:
            bid = self._respond_to_one_suit_no_support()
            if bid.is_pass:
                bid = self._minor_opening_check_support()
        else:
            bid = self._respond_to_minor()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _minor_opening_check_support(self):
        """Respond after one of minor, check support."""
        if self.suit_length(self.opener_bid_one.denomination) >= 4:
            bid = self._respond_to_one_suit_with_support()
        else:
            bid = self._respond_to_one_suit_no_support()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_minor(self):
        """Return bid after opener has bid minor suit."""
        hand_value_points = self.support_points(self.opener_bid_one.denomination)
        if hand_value_points <= 9:
            minor_level = 2
        elif hand_value_points <= 10:
            minor_level = 3
        elif hand_value_points <= 13:
            minor_level = 4
        else:
            minor_level = 5
        if hand_value_points <= 4:
            bid = Pass('0075')
        elif (self.hcp >= 16 and
              self.second_suit == self.opener_bid_one.denomination and
              self.longest_suit not in self.opponents_suits):
            bid = self.next_level_bid(self.longest_suit, '0781', raise_level=1)
        elif self._is_five_four_game_going():
            bid = self.next_level_bid(self.second_suit, '0076', raise_level=1)
        elif self._is_balanced_with_no_long_minor():
            bid = self._nt_response_to_minor()
        elif self._has_six_card_unbid_suit():
            if self.hcp >= 16:
                raise_level = 1
            else:
                raise_level = 0
            bid = self.next_level_bid(self.longest_suit, '0077', raise_level)
        elif self._is_semi_balanced_some_support():
            bid = self.nt_bid(2, '0078')
        elif self.opener_bid_one.denomination == self.club_suit and minor_level > 0:
            bid = self._respond_to_clubs(minor_level)
        elif self.opener_bid_one.denomination == self.diamond_suit and minor_level > 0:
            bid = self._response_to_diamonds(minor_level)
        else:
            bid = self._respond_to_one_suit_no_support()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _nt_response_to_minor(self):
        """Bid NT in response to minor opening."""
        if self.hcp >= 16 and self.longest_suit not in self.opponents_suits:
            bid = self.next_level_bid(self.longest_suit, '0000', raise_level=1)
        elif self.hcp >= 13 and self.nt_level <= 3:
            bid = self.nt_bid(3, '0079')
        elif self.hcp >= 10 and self.nt_level <= 2:
            bid = self.nt_bid(2, '0080')
        elif self.hcp >= 6 and self.nt_level <= 1:
            bid = self.nt_bid(1, '0081')
        else:
            bid = Pass('0082')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_clubs(self, level):
        """Respond after club opening."""
        club_level = self.next_level(self.club_suit)
        hand_value_points = self.support_points(self.opener_bid_one.denomination)
        if level > 3 and self.hcp >= 13 and self.nt_level <= 3 and self.shape[1] <= 4:
            bid = self.nt_bid(3, '0083')
        elif self._has_five_card_club_support_and_ten_points():
            if self.hcp >= 15:
                level = 5
            elif self.hcp >= 12:
                level = 4
            else:
                level = 3
            bid = self.suit_bid(level, self.club_suit, '0249')
        elif self._has_five_card_suit_and_ten_points():
            bid = self.next_level_bid(self.longest_suit, '0084')
        elif (self.clubs >= 4 and
              hand_value_points >= 5 and
              club_level <= level):
            bid = self.club_bid(level, '0085', True)
        elif self.diamonds >= 5:
            bid = self._respond_to_clubs_with_diamonds()
        elif self.clubs >= 4:
            bid = self._respond_to_one_suit_with_support()
        else:
            bid = self._respond_to_one_suit_no_support()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_clubs_with_diamonds(self):
        """Respond after club opening with diamonds."""
        if self.diamond_suit in self.opponents_suits:
            bid = Pass('0522')
        elif self.hcp + self.shape[0] >= 16:
            bid = self.next_level_bid(self.diamond_suit, '0086')
        elif self._can_rebid_diamonds_at_next_level():
            bid = self.next_level_bid(self.diamond_suit, '0087')
        else:
            bid = Pass('0088')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _response_to_diamonds(self, level):
        """Respond after diamond opening."""
        club_level = self.next_level(self.club_suit)
        diamond_level = self.next_level(self.diamond_suit)
        hand_value_points = self.support_points(self.opener_bid_one.denomination)
        if 6 <= self.hcp <= 9 and self.is_balanced and self.stoppers_in_bid_suits and self.nt_level == 1:
            bid = self.nt_bid(1, '0312')
        elif self.is_balanced and 10 <= hand_value_points <= 12:
            bid = self.diamond_bid(3, '0764', True)
        elif self.diamonds >= 4 and 10 <= hand_value_points <= 14:
            level = max(level, diamond_level)
            bid = self.diamond_bid(level, '0089', True)
        elif self.clubs >= 5 and hand_value_points >= 9 and club_level <= 2:
            level = 2
            if self.hcp >= 16:
                level = 3
            bid = self.club_bid(level, '0090', True)
        elif self.diamonds >= 4:
            bid = self._respond_to_one_suit_with_support()
        else:
            bid = self._respond_to_one_suit_no_support()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_one_suit_with_support(self):
        """Respond with support for opener."""
        suit = self.opener_bid_one.denomination
        level = self.next_level(suit)
        hand_value_points = (self.hcp
                             + self.support_shape_points(suit)
                             + self.suit_length(suit) - 4)

        if suit.is_minor:
            bid = self._respond_to_one_minor_with_support()
        elif hand_value_points < 6:
            bid = Pass('0091', True)
        elif (6 <= hand_value_points <= 9 and
              level <= self.opener_bid_one.level + 1):
            bid = self.suit_bid(level, suit, '0092', True)
        elif (10 <= hand_value_points <= 11 and
              level <= self.opener_bid_one.level + 2):
            bid = self.suit_bid(level + 1, suit, '0094', True)
        elif 12 <= hand_value_points <= 15 and level <= self.opener_bid_one.level + 3:
            if self.suit_length(suit) >= 5 and self.hcp >= 11:
                level += (self.suit_length(suit) - 4)
            if 10 <= self.hcp <= 11:
                level -= 1
            level = min(2, level)
            bid = self.suit_bid(level + 2, suit, '0094', True)
        elif suit.is_major and hand_value_points >= 18:
            bid = self.nt_bid(4, '0095', True)  # Blackwood
        elif self.hcp >= 16 and suit.is_minor:
            bid = self.suit_bid(6, suit, '0339')
        elif self.hcp >= 16 and self.shape[1] >= 4 and self.longest_suit not in self.opponents_suits:
            if suit == self.longest_suit and self.second_suit not in self.opponents_suits:
                suit = self.second_suit
            else:
                suit = self.longest_suit
            bid = self.next_level_bid(suit, '0479', raise_level=1)
        elif hand_value_points >= 13:
            bid = self.bid_to_game(suit, '0096')
        else:
            bid = Pass('0097', True)
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_one_minor_with_support(self):
        """Respond with support for opener."""
        suit = self.opener_bid_one.denomination
        level = self.next_level(suit)
        hand_value_points = (self.hcp
                             + self.support_shape_points(suit)
                             + self.suit_length(suit) - 4)
        if hand_value_points < 6:
            bid = Pass('0091', True)
        elif (6 <= hand_value_points <= 9 and
              level <= self.opener_bid_one.level + 1):
            bid = self.suit_bid(level, suit, '0877', True)
        elif (10 <= hand_value_points <= 11 and
              level <= self.opener_bid_one.level + 2):
            bid = self.suit_bid(level + 1, suit, '0093', True)
        elif 12 <= hand_value_points <= 15 and level <= self.opener_bid_one.level + 3:
            if self.suit_length(suit) >= 5 and self.hcp >= 11:
                level += (self.suit_length(suit) - 4)
            if 10 <= self.hcp <= 11:
                level -= 1
            level = min(2, level)
            bid = self.suit_bid(level + 2, suit, '0094', True)
        elif suit.is_major and hand_value_points >= 18:
            bid = self.nt_bid(4, '0095', True)  # Blackwood
        elif self.hcp >= 16 and suit.is_minor:
            bid = self.suit_bid(6, suit, '0339')
        elif self.hcp >= 16 and self.shape[1] >= 4 and self.longest_suit not in self.opponents_suits:
            if suit == self.longest_suit and self.second_suit not in self.opponents_suits:
                suit = self.second_suit
            else:
                suit = self.longest_suit
            bid = self.next_level_bid(suit, '0479', raise_level=1)
        elif hand_value_points >= 13:
            bid = self.bid_to_game(suit, '0096')
        else:
            bid = Pass('0097', True)
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_one_suit_no_support(self):
        """Respond to opening suit bid with no support."""
        own_suit = self._suit_for_response_no_support()
        response_level = self._level_for_response(own_suit)
        if own_suit == self.no_trumps:
            bid = self._no_support_bid_nt()
        elif self.hcp >= 16:
            bid = self._no_support_strong(own_suit, response_level)
        elif 9 <= self.hcp <= 15:
            bid = self._no_support_intermediate(response_level, own_suit, '0098')
        elif 6 <= self.hcp <= 8:
            bid = self._no_support_weak(response_level, own_suit)
        elif self.hcp == 5 and response_level == 1 and self.shape[0] >= 5:
            bid = self.suit_bid(1, own_suit, '0099')
        else:  # elif self.hcp <= 5:
            bid = Pass('0100')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _no_support_bid_nt(self):
        """Respond to opening suit bid with no support, nt."""
        if self.hcp >= 8:
            stoppers = self.four_in_bid_suits()
        else:
            stoppers = self.stoppers_in_bid_suits
        poor_stoppers = self.poor_stoppers_in_bid_suits
        hold_opponents_suit = False
        if (self.overcaller_bid_one.is_suit_call and
                self.suit_length(self.overcaller_bid_one.denomination) >= 5):
            hold_opponents_suit = True
        if self.hcp <= 9 and hold_opponents_suit:
            bid = Pass('0101')
        elif stoppers or (self.shape[0] == 4 and poor_stoppers):
            bid = self._bid_nt_stoppers()
        else:
            bid = Pass('0102')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _bid_nt_stoppers(self):
        """Return NT bid with stoppers."""
        if self.nt_level <= 3 and self.hcp >= 13:
            bid = self.nt_bid(3, '0103')
        elif self.nt_level <= 2 and self.hcp >= 10:
            bid = self.nt_bid(2, '0104')
        elif self.nt_level == 1 and self.hcp >= 6:
            bid = self.nt_bid(1, '0105')
        else:
            bid = Pass('0106')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _no_support_strong(self, own_suit, response_level):
        """Respond to opening suit bid with no support, strong."""
        level = self._get_level__no_support_strong(own_suit, response_level)
        bid = self.suit_bid(level, own_suit, '0107')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _no_support_intermediate(self, response_level, own_suit, comment=''):
        """Return bid with no support for opener's suit, intermediate."""
        trial_bid = self._no_support_intermediate_trial_bid(response_level, own_suit, comment)
        if (self.opener_bid_one.name == '1S' and
                trial_bid.name == '2H' and self.hearts < 5):
            bid = self._no_support_intermediate_after_spade()
        elif self._has_five_card_major_and_game_going():
            bid = self.next_level_bid(self.five_card_major_suit, '0098')
        elif self._is_balanced_and_strong():
            bid = self.nt_bid(3, '0136')
        elif self._is_game_going_support_for_partners_minor(response_level):
            bid = Double('0408')
        elif (self._has_four_card_major_and_game_going(response_level) and
              self.longest_suit == self.opener_bid_one.denomination):
            suit = self.next_four_card_suit()
            bid = self.next_level_bid(suit, '0108')
        elif self._has_overcall_but_no_support(own_suit):
            if self.hcp >= 14:
                level = 3
            else:
                level = 2
            bid = self.nt_bid(level, '0109')
        else:
            bid = self._no_support_intermediate_not_spade(own_suit, trial_bid)
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _no_support_intermediate_trial_bid(self, level, own_suit, comment):
        """Return trial bid with no support, intermediate."""
        if level == 1 and not comment:
            comment = '0110'
        elif own_suit == self.opener_bid_one.denomination:
            comment = '0093'
        else:
            comment = '0111'
        bid = self.suit_bid(level, own_suit, comment)
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _no_support_intermediate_after_spade(self):
        """Return bid with no support for opener's spade."""
        if self._has_good_diamonds():
            bid = self.diamond_bid(2, '0112')
        elif self._has_good_clubs():
            bid = self.club_bid(2, '0113')
        else:
            hcp = self.hcp
            level = self.quantitative_raise(hcp, 0, [0, 10, 13], 3)
            if level >= self.nt_level:
                bid = self.nt_bid(level, '0114')
            else:
                bid = Pass('0115')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _no_support_intermediate_not_spade(self, own_suit, trial_bid):
        """Return bid with no support for opener's suit, intermediate."""
        if own_suit == self.spade_suit and self.hcp >= 10:
            bid = trial_bid
            bid.comment = '0116'
        elif own_suit.is_minor and self.is_balanced and self.hcp == 9:
            bid = self._no_support_intermediate_minor_balanced()
        elif self._is_semi_balanced_no_support(own_suit) and self.hcp <= 12:
            bid = self._no_support_intermediate_minor_balanced_overcall()
        elif self._four_four_one_stoppers():
            bid = self._singleton_in_openers_suit(trial_bid)
        elif self._has_singleton_in_openers_suit_and_stoppers(own_suit):
            bid = self.nt_bid(2, '0117')
        elif self._has_no_support_and_no_long_suit(own_suit, trial_bid):
            if self.hcp >= 13:
                level = 3
                comment = '0161'
            else:
                comment = '0118'
                level = 2
            bid = self.nt_bid(level, comment)
        elif self._forced_to_jump_bid(own_suit, trial_bid):
            bid = Pass('0119')
        else:
            bid = trial_bid
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _singleton_in_openers_suit(self, trial_bid):
        """Return bid if holding a singleton in openers suit."""
        if self.nt_level == 2:
            if self.overcaller_bid_one.is_nt:
                if self.hcp >= 10:
                    bid = self.nt_bid(2, '0120')
                else:
                    bid = Pass('0121')
            else:
                bid = self.nt_bid(2, '0122')
        else:
            bid = trial_bid
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _no_support_intermediate_minor_balanced(self):
        """Return bid with no support balanced but minor suit."""
        if self.nt_level == 1 and self.stoppers_in_bid_suits:
            if self.shape[1] == 4 and self.next_level(self.longest_suit) == 1:
                bid = self.next_level_bid(self.longest_suit, '0123')
            else:
                bid = self.nt_bid(1, '0124')
        else:
            bid = Pass('0125')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _no_support_intermediate_minor_balanced_overcall(self):
        """Return bid with no support balanced, overcall, minor suit."""
        if (self.overcaller_bid_one.name == '1NT' and
                self.hcp >= 10):
            bid = Double('0126')
        else:
            if self.hcp >= 13:
                level = 3
            elif self.hcp >= 10:
                level = 2
            else:
                level = self.nt_level
            if level >= self.nt_level:
                bid = self.nt_bid(level, '0127')
            else:
                bid = Pass('0128')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _no_support_weak(self, response_level, own_suit):
        """Respond to opening suit bid with no support, weak."""
        if response_level == 1:
            bid = self._no_support_weak_level_one(response_level, own_suit)
        else:
            bid = self._no_support_weak_over_level_one(own_suit)
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _no_support_weak_level_one(self, level, own_suit):
        """Respond to bid with no support at level one, weak."""
        if self.next_level(own_suit) == 1:
            bid = self.suit_bid(level, own_suit, '0129')
        else:
            suit = self.second_suit
            level = self.next_level(suit)
            bid = self.suit_bid(level, suit, '0130')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _no_support_weak_over_level_one(self, own_suit):
        """Respond to bid with no support above level one, weak."""
        major = self.four_card_major_suit
        if major and self.next_level(major) == 1:
            bid = self.next_level_bid(major, '0131')
        elif (not self.overcaller_bid_one.is_pass and
              not self.overcaller_bid_one.is_double):
            bid = self._no_support_weak_over_level_one_overcall()
        elif own_suit.is_minor:
            bid = self._no_support_weak_over_level_one_minor()
        else:
            bid = self.nt_bid(1, '0132')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _no_support_weak_over_level_one_overcall(self):
        """Respond to bid with no support above level one, overcall."""
        stopper = False
        if not self.overcaller_bid_one.is_nt:
            stopper = self.has_stopper(self.overcaller_bid_one.denomination)
        if stopper and self.nt_level <= 1:
            bid = self.nt_bid(1, '0133')
        elif (self.shape[0] >= 7 and
              self.hcp >= 8 and
              self.longest_suit not in self.opponents_suits):
            bid = self.next_level_bid(self.longest_suit, '0134')
        else:
            bid = Pass('0135')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _no_support_weak_over_level_one_minor(self):
        """Respond to bid with no support above level one, minor."""
        if (self.hearts >= 4 and
                self.next_level(self.heart_suit) == 1):
            bid = self.heart_bid(1, '0136')
        elif (self.spades >= 4 and
              self.next_level(self.spade_suit) == 1):
            bid = self.spade_bid(1, '0137')
        elif self.shape[0] >= 8 and self.hcp >= 8 and self.next_level(self.longest_suit) <= 2:
            bid = self.next_level_bid(self.longest_suit, '0783')
        else:
            bid = self.nt_bid(1, '0138')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_two_clubs(self):
        """Rebid after 2C opening."""
        suit = self._get_suit_after_two_clubs_opening()
        quality = HandSuit(suit, self).suit_quality()
        suit_length = self.suit_length(self.longest_suit)
        if ((self.hcp > 7 or
             (self.hcp == 7 and
              quality >= 2)) and suit_length >= 5):
            bid = self._suit_response_after_two_clubs(suit)
        elif self.is_balanced and self.hcp > 7:
            bid = self.nt_bid(2, '0139')
        elif (self.diamond_suit not in self.opponents_suits and
              self.next_level(self.diamond_suit) <= 2):
            bid = self.diamond_bid(2, '0140')
        else:
            bid = Pass('0254')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _suit_response_after_two_clubs(self, own_suit):
        """Make suit response after 2C."""
        if own_suit == self.spade_suit:
            bid = self.spade_bid(2, '0141')
        elif own_suit == self.heart_suit:
            bid = self.heart_bid(2, '0142')
        elif own_suit == self.diamond_suit:
            bid = self.diamond_bid(3, '0143')
        else:  # elif own_suit == self.club_suit:
            bid = self.club_bid(3, '0144')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_weak_two(self):
        """Respond to weak two."""
        openers_suit = self.opener_bid_one.denomination
        hand_value_points = (self.hcp +
                             self.support_shape_points(openers_suit))
        if self.hcp <= 10:
            bid = self._respond_to_weak_two_weak()
        elif self.overcaller_bid_one.is_double:
            bid = Pass('0145')
        elif self._support_after_suit_overcall_game_going():
            level = self.next_level(openers_suit)
            bid = self.suit_bid(level, openers_suit, '0146', True)
        elif hand_value_points >= 16:
            bid = self._respond_to_weak_two_strong()
        elif self._has_singleton_in_openers_suit_and_own_six_card_suit():
            bid = self.next_level_bid(self.longest_suit, '0147')
        else:
            bid = Pass('0148')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_weak_two_weak(self):
        """Respond to weak two if weak."""
        openers_suit = self.opener_bid_one.denomination
        level = self.next_level(openers_suit)
        if self.suit_length(openers_suit) >= 4 and level <= 4:
            bid = self.suit_bid(4, openers_suit, '0149')
        elif self.suit_length(openers_suit) >= 3 and level <= 3:
            bid = self.suit_bid(3, openers_suit, '0150')
        else:
            bid = Pass('0151')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_weak_two_strong(self):
        """Respond to weak two if strong."""
        if self._has_two_card_support_for_major():
            bid = self.suit_bid(4, self.opener_bid_one.denomination, '0152', True)
        elif self.hcp >= 16 and self._shape[0] >=8 and self.longest_suit.is_major:
            bid = self.bid_to_game(self.longest_suit, '0427')
        elif self._has_own_suit():
            bid = self.next_level_bid(self.longest_suit, '0153')
        elif self.hcp >= 23 and self.suit_length(self.opener_bid_one.denomination) >= 2 and self.is_balanced:
            bid = self.nt_bid(3, '0905')
        elif self.hcp >= 23 and self.suit_length(self.opener_bid_one.denomination) >= 2:
            bid = self.suit_bid(6, self.opener_bid_one.denomination, '0622')
        elif (self.hcp >=15 and
                self.shape[1] >= 5 and
                self.longest_suit.is_major and
                self.second_suit.is_major and
                self.longest_suit not in self.opponents_suits):
            bid = self.next_level_bid(self.longest_suit, '0921')
        else:
            if self.nt_level <= 3:
                bid = self.nt_bid(3, '0154')
            else:
                bid = Pass('0000')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_weak_three(self):
        """Respond to weak three."""
        if self.hcp >= 15:
            bid = self._respond_to_weak_three_strong()
        elif (self.hcp <= 9 and
              self.suit_length(self.opener_bid_one.denomination) >= 2):
            bid = self.suit_bid(4, self.opener_bid_one.denomination, '0155')
        else:
            bid = Pass('0156')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_to_weak_three_strong(self):
        """Respond to weak three, strong."""
        if self._two_card_support_for_major_fifteen_points():
            bid = self.suit_bid(4, self.opener_bid_one.denomination, '0157')
        elif self.is_balanced and self.nt_level <= 3:
            bid = self.nt_bid(3, '0158')
        elif (self.opener_bid_one.denomination.is_major and
              self.suit_length(self.opener_bid_one.denomination) >= 1):
            bid = self.suit_bid(4, self.opener_bid_one.denomination, '0159')
        elif self._can_bid_six_card_suit_at_level_three():
            bid = self.next_level_bid(self.longest_suit, '0832')
        elif self.nt_level <= 3:
            bid = self.nt_bid(3, '0160')
        else:
            bid = Pass('0161')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    # Various utility functions

    def _get_suit_after_two_clubs_opening(self):
        """Determine best suit after 2C opening."""
        if self.five_five:
            if self.longest_suit < self.second_suit:
                suit = self.second_suit
            else:
                suit = self.longest_suit
        else:
            suit = self.longest_suit
        return suit

    def _get_one_nt_jump_level(self):
        """Return jump level for response after 1NT with overcall."""
        if self.hcp >= 12:
            jump_level = 1
        else:
            jump_level = 0
        return jump_level

    def _level_for_response(self, own_suit):
        """Return level for no support bid."""
        level = self.next_level(own_suit)
        if self._has_overcall_ten_points_shortage(own_suit):
            level += 1
        elif self._has_support_nine_points_shortage(own_suit):
            level += 1
        return level

    def _get_level__no_support_strong(self, own_suit, level):
        """Return level for strong bid with no support."""
        if self.longest_suit.is_major and self.shape[0] >= 6:
            level += 2
        else:
            level += 1
        jump_level = self.next_level(own_suit) + 1
        level = min(level, jump_level)
        return level

    def _suit_for_response_no_support(self):
        """Return suit for responder when no support for opener."""
        if self.five_five_or_better:
            next_suit = self._higher_five_card_suit()
        else:
            next_suit = self._cheapest_four_card_suit(self.opener_bid_one.denomination)
        suit = self.longest_suit
        if suit in self.opponents_suits:
            suit = self.second_suit
        if self.suit_length(suit) == 4:
            for bid in self.bid_history[::-1]:
                if Bid(bid).is_value_call:
                    suit = self._cheapest_four_card_suit(Bid(bid).denomination)
                    break
        if self.suit_length(suit) < 4:
            suit = self.no_trumps
        elif suit == self.overcaller_bid_one.denomination:
            suit = self.no_trumps
        elif (self.hcp >= 16 and
              self.longest_suit == self.opener_bid_one.denomination and
              self.shape[1] == 4 and
              next_suit and
              next_suit not in self.opponents_suits):
            suit = next_suit
        return suit

    def _cheapest_four_card_suit(self, bid_suit):
        """Return cheapest 4 card suit after given suit."""
        loop = 0
        rank = bid_suit.rank
        while True:
            rank = (rank + 1) % 4
            if self.suit_length(self.suits[rank]) >= 4:
                break
            loop += 1
            if loop > 4:
                rank = 4
                break
        suit_name = ['C', 'D', 'H', 'S', None][rank]
        suit = Suit(suit_name)
        return suit

    def _higher_five_card_suit(self):
        """Return highest five card suit."""
        for rank in list(range(4))[::-1]:
            if self.suit_length(self.suits[rank]) >= 5:
                suit_name = ['C', 'D', 'H', 'S'][rank]
                suit = Suit(suit_name)
                return suit

    def _get_best_major(self):
        """Return the best major suit."""
        if self.spades >= 6:
            suit = self.spade_suit
        elif self.hearts >= 6:
            suit = self.heart_suit
        elif self.spades == 5:
            suit = self.spade_suit
        elif self.hearts == 5:
            suit = self.heart_suit
        elif self.hearts == 4:
            suit = self.heart_suit
        else:
            suit = self.spade_suit
        return suit

    def _one_nt_max_level(self):
        """Return maximum level for response after 1NT with overcall."""
        if (self.hcp >= 12 or
                (self.hcp == 11 and self.shape[0] >= 6)):
            max_level = 3
        elif self.hcp >= 8:
            max_level = 2
        elif self.hcp >= 6 and self.shape[0] >= 6:
            max_level = 2
        else:
            max_level = 1
        return max_level

    def _trial_bid_suit_after_one_nt_with_overcall(self):
        """Try a suit after 1NT and overcall."""
        jump_level = self._get_one_nt_jump_level()
        trial_bid = self.next_level_bid(self.longest_suit, '0039', jump_level)
        return trial_bid

    # Various boolean functions

    def _six_card_major_thirteen_points(self):
        """Return True if hand has a six+ card_major with 13 points."""
        result = (self.shape[0] >= 6 and
                  self.longest_suit not in self.opponents_suits and
                  self.longest_suit.is_major and
                  self.hcp >= 13)
        return result

    def _can_bid_after_one_nt_with_overcall(self):
        """Return True if suit bid can be made after 1NT with overcall."""
        max_level = self._one_nt_max_level()
        trial_bid = self._trial_bid_suit_after_one_nt_with_overcall()
        result = (self.shape[0] >= 5 and
                  trial_bid.denomination not in self.opponents_suits and
                  trial_bid.level <= max_level)
        return result

    def _is_semi_balanced_eleven_plus(self):
        """Return True if semi balanced and 11+ points and stoppers."""
        result = False
        if self.is_semi_balanced and self.hcp >= 11:
            if self.stoppers_in_bid_suits:
                result = True
            elif (self.suit_length(self.overcaller_bid_one.denomination) >= 2 and
                  self.suit_points(self.overcaller_bid_one.denomination) >= 2):
                result = True
        return result

    def _is_balanced_game_going(self):
        """Return True if balanced with 13+ points."""
        result = (self.hcp >= 13 and
                  self.is_balanced and
                  # self.stoppers_in_bid_suits and
                  self.nt_level <= 3)
        return result

    def _is_game_going(self):
        """Return True with 13+ points."""
        result = (self.hcp >= 13 and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _is_five_four_game_going(self):
        """Return True with 5/4 and 14+ points."""
        result = (self.hcp >= 14 and
                  self.five_four and
                  self.second_suit not in self.opponents_suits)
        return result

    def _is_balanced_with_no_long_minor(self):
        """Return True if balanced with only 4 card minors."""
        result = (self.is_balanced and
                  self.clubs < 5 and self.diamonds < 5 and
                  self.stoppers_in_bid_suits)
        return result

    def _has_six_card_unbid_suit(self):
        """Return True with 6+ card unbid suit."""
        result = (self.shape[0] >= 6 and
                  self.hcp >= 9 and
                  self.opener_bid_one.denomination != self.longest_suit
                  and self.longest_suit not in self.opponents_suits)
        return result

    def _is_semi_balanced_some_support(self):
        """Return True if semi balanced with 3+cards in partners suit."""
        result = (self.is_semi_balanced and
                  self.suit_length(self.opener_bid_one.denomination) >= 3 and
                  self.stoppers_in_bid_suits and
                  11 <= self.hcp <= 12 and
                  self.longest_suit == self.opener_bid_one.denomination and
                  self.nt_level <= 2 and
                  self.shape[3] >= 2)
        return result

    def _has_five_card_club_support_and_ten_points(self):
        """Return True with 5+ support for clubs and 10+ points."""
        result = (self.longest_suit == self.club_suit and
                  self.shape[0] >= 5 and
                  self.hcp >= 10 and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _has_five_card_suit_and_ten_points(self):
        """Return True with 5 card suits and 10+ points."""
        result = (self.shape[0] >= 5 and
                  10 <= self.hcp <= 15 and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _is_balanced_fewer_than_eight_points(self):
        """Return True if balanced and fewer than 8 points."""
        result = (6 <= self.hcp <= 8 and
                  self.is_balanced and
                  self.stoppers_in_bid_suits and
                  self.nt_level == 1)
        return result

    def _has_overcall_ten_points_shortage(self, own_suit):
        """Return True if overcall and 10 points and doubleton."""
        result = (self.overcall_made and self.hcp >= 10 and
                  self.shape[3] <= 2 and
                  self.shape[self.opener_bid_one.denomination.rank] >= 2 and
                  own_suit == self.opener_bid_one.denomination)
        return result

    def _has_support_nine_points_shortage(self, own_suit):
        """Return True if 5 card support and 9 points and singleton."""
        result = (own_suit == self.opener_bid_one.denomination and
                  self.suit_length(own_suit) >= 5 and
                  self.shape[3] <= 1 and
                  self.hcp >= 9)
        return result

    def _has_five_card_major_and_game_going(self):
        """Return True with 5 card major and 13+ points."""
        result = (self.five_card_major_suit and
                  self.hcp >= 13 and
                  self.five_card_major_suit not in self.opponents_suits)
        return result

    def _is_balanced_and_strong(self):
        """Return True if balanced and 15+ points."""
        result = (self.hcp >= 15 and
                  self.shape[0] == 4 and
                  self.stoppers_in_bid_suits and
                  self.nt_level <= 3 and
                  (not self.four_card_major_or_better or
                   self.overcall_made))
        return result

    def _has_four_card_major_and_game_going(self, response_level):
        """Return True with 4 card major and 13+ points."""
        result = (self.four_card_major_suit and
                  response_level >= 2 and
                  self.hcp >= 13 and
                  self.four_card_major_suit not in self.opponents_suits and
                  (self.opener_bid_one.denomination != self.spade_suit or
                   self.four_card_major_suit != self.heart_suit)  # and
                  # self.suit_length(self.opener_bid_one.denomination) >= 4
                  )
        return result

    def _has_overcall_but_no_support(self, suit):
        """Return True if no support with overcall."""
        result = (self.nt_level == 2 and
                  self.suit_length(suit) <= 4 and
                  suit != self.opener_bid_one.denomination and
                  self.overcaller_bid_one.is_suit_call
                  and self.stoppers_in_bid_suits)
        return result

    def _has_good_diamonds(self):
        """Return True with good diamonds."""
        diamond_level = self.next_level(self.diamond_suit)
        result = (diamond_level <= 2 and (self.diamonds >= 5 or
                                          (self.diamonds == 4 and
                                           self.hcp >= 10)))
        return result

    def _has_good_clubs(self):
        """Return True with good clubs."""
        club_level = self.next_level(self.club_suit)
        result = (club_level <= 2 and (self.clubs >= 5 or
                                       (self.clubs == 4 and
                                        self.hcp >= 10)))
        return result

    def _is_semi_balanced_no_support(self, own_suit):
        """Return True if semi balanced and overcall."""
        result = (self.stoppers_in_bid_suits and
                  self.is_semi_balanced and
                  own_suit.is_minor and
                  self.overcall_made and
                  (own_suit != self.opener_bid_one.denomination or
                   self.hcp >= 10))
        return result

    def _four_four_one_stoppers(self):
        """Return True if 4441 and singleton in opener's suit."""
        result = (self.stoppers_in_bid_suits and
                  self.shape[0] == 4 and
                  self.suit_length(self.opener_bid_one.denomination) == 1)
        return result

    def _has_singleton_in_openers_suit_and_stoppers(self, own_suit):
        """Return True if singleton in opener's suit and stopper, level 2."""
        result = (self.stoppers_in_bid_suits and
                  self.overcall_made and
                  self.suit_length(self.opener_bid_one.denomination) == 1
                  and own_suit != self.longest_suit and
                  self.hcp >= 10 and self.nt_level == 2)
        return result

    def _has_no_support_and_no_long_suit(self, own_suit, trial_bid):
        """Return True with no support and no long suit."""
        result = (trial_bid.level >= 2 and
                  self.suit_length(own_suit) < 5 and
                  self.is_jump(self.opener_bid_one, trial_bid) and
                  own_suit != self.opener_bid_one.denomination and
                  self.is_balanced and
                  self.hcp >= 10 and self.nt_level <= 2 and
                  (self.stoppers_in_bid_suits or
                   (self.is_balanced and self.shape[3] >= 3)))
        return result

    def _forced_to_jump_bid(self, own_suit, trial_bid):
        """Return True if forced to jump."""
        result = (trial_bid.level >= 2 and
                  self.suit_length(own_suit) < 5 and
                  self.is_jump(self.opener_bid_one, trial_bid) and
                  self.hcp <= 10)
        return result

    def _support_after_suit_overcall_game_going(self):
        """Return True with support after overcall."""
        hand_value_points = (self.hcp +
                             self.support_shape_points(self.opener_bid_one.denomination))
        result = (12 <= hand_value_points <= 15 and
                  self.suit_length(self.opener_bid_one.denomination) >= 2 and
                  self.overcaller_bid_one.is_value_call)
        return result

    def _has_singleton_in_openers_suit_and_own_six_card_suit(self):
        """Return True if singleton in opener's suit and a 6+card suit."""
        result = (self.shape[0] >= 6 and
                  self.hcp >= 12 and
                  self.suit_length(self.opener_bid_one.denomination) <= 1 and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _has_two_card_support_for_major(self):
        """Return True if openers suit is major and two card support."""
        result = (self.opener_bid_one.denomination.is_major and
                  self.suit_length(self.opener_bid_one.denomination) >= 2)
        return result

    def _has_own_suit(self):
        if self.shape[0] >= 6:
            suit_quality = HandSuit(self.longest_suit, self).suit_quality()
            has_own_suit = True
        else:
            suit_quality = 0
            has_own_suit = False
        result = (has_own_suit and suit_quality >= 2 and
                  self.suit_length(self.opener_bid_one.denomination) <= 1 and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _two_card_support_for_major_fifteen_points(self):
        """Return True with support for opener's major."""
        result = (self.opener_bid_one.denomination.is_major and
                  self.suit_length(self.opener_bid_one.denomination) >= 2 and
                  self.hcp >= 15 and
                  self.next_level(self.opener_bid_one.denomination) <= 4)
        return result

    def _is_game_going_support_for_partners_minor(self, response_level):
        """Return True if game going and support for openers minor."""
        result = (self._has_four_card_major_and_game_going(response_level) and
                  self.next_four_card_suit() == self.opener_bid_one.denomination and
                  self.opener_bid_one.denomination.is_minor)
        return result

    def _can_rebid_diamonds_at_next_level(self):
        """Return True if can bid diamonds at the next level."""
        hand_value_points = self.support_points(self.opener_bid_one.denomination)
        level = self.next_level(self.diamond_suit)
        result = ((hand_value_points >= 5 and level <= 1) or
                  (hand_value_points >= 9 and
                   self.suit_points(self.diamond_suit) >= 5 and
                   level <= 2))
        return result

    def _can_bid_six_card_suit_at_level_three(self):
        """Return True if Can bid six card suit at 3 level."""
        result = (self.shape[0] >= 6 and
                  self.next_level(self.longest_suit) <= 3 and
                  self.longest_suit not in self.opponents_suits)
        return result

    # def xxx(self):
    #     """Return True if xxx."""
    #     result = False
    #     return result

    # def xxx(self):
    #     """Return True if xxx."""
    #     result = False
    #     return result



""" Bid for Game
    Acol Blackwood module
"""

import inspect
from .bidding import Bid, Pass
from .hand import Hand


class Blackwood(Hand):
    """ Return BfG Blackwood class."""
    HONOURS = {'aces': 1, 'kings': 2}

    def __init__(self, hand_cards, board):
        super(Blackwood, self).__init__(hand_cards, board)

        self.trace = 0

        self.my_last_bid = Bid(self.bid_history[-4], '')
        self.partners_last_bid = Bid(self.bid_history[-2], '')
        if len(self.bid_history) >= 7:
            self.partners_penultimate_bid = Bid(self.bid_history[-6], '')
            self.responders_second_bid = Bid(self.bid_history[6], '')

        self.openers_first_bid = Bid(self.bid_history[0], '')
        self.responders_first_bid = Bid(self.bid_history[2], '')
        if len(self.bid_history) >= 5:
            self.openers_second_bid = Bid(self.bid_history[4], '')
        if len(self.bid_history) >= 9:
            self.my_penultimate_bid = Bid(self.bid_history[-8], '')
            self.opener_bid_three = Bid(self.bid_history[8], '')
        if len(self.bid_history) >= 11:
            self.responders_third_bid = Bid(self.bid_history[10], '')
        self.ace_count = self.count_aces()
        self.king_count = self.count_kings()
        self.four_aces = self.has_four_aces()
        self.three_aces = self.has_three_aces()

    def count_aces(self):
        """Return bid for Aces in hand."""
        bid = self._count_honours(self.HONOURS['aces'])
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def count_kings(self):
        """Return bid for Kings in hand."""
        bid = self._count_honours(self.HONOURS['kings'])
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _count_honours(self, honour):
        """Return bid based on number of aces or kings in hand."""
        if honour == self.HONOURS['aces']:
            honours = self.aces
            bid_level = 5
            comment = '0864' # don't change
        else:
            honours = self.kings
            bid_level = 6
            comment = '0865'
        suit_name = [self.club_suit, self.diamond_suit, self.heart_suit,
                     self.spade_suit, self.club_suit][honours]
        level = self.next_level(suit_name)
        if level <= bid_level:
            bid = self.suit_bid(bid_level, suit_name, comment)
            if bid.name == '6C' and self.kings == 4:
                bid = self.nt_bid(6, '0939')
        else:
            bid = Pass('0866')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def has_four_aces(self):
        """Return True if partnership has 4 Aces."""
        suit = self.partners_last_bid.denomination
        aces = 0
        if suit.is_suit:
            partners_aces = suit.rank
            if self.aces == 0 and partners_aces == 0:
                partners_aces = 4
            aces = self.aces + partners_aces
        value = aces == 4
        self.tracer(__name__, inspect.currentframe(), value, self.trace)
        return value

    def has_three_aces(self):
        """Return True if partnership has 3 Aces."""
        suit = self.partners_last_bid.denomination
        aces = 0
        if suit.is_suit:
            partners_aces = suit.rank
            if self.aces == 0 and partners_aces == 0:
                partners_aces = 4
            aces = self.aces + partners_aces
        value = aces >= 3
        self.tracer(__name__, inspect.currentframe(), value, self.trace)
        return value

    def select_contract(self):
        """Return the selected contract after Blackwood process."""
        suit = self.select_contract_suit()
        aces = self.partnership_aces()
        kings = self.partnership_kings()
        if aces == 4 and kings == 4 and (suit.is_major or suit.is_nt) :
            level = 7
        elif aces == 4 and kings == 3:
            level = 6
        elif self.hcp < 20 and self.queens+self.jacks+self.tens+self.nines <= 4:
            level = 5
        elif aces < 3:
            level = 5
        else:
            level = 6
        partners_last_bid = Bid(self.bid_history[-2])
        if (partners_last_bid.denomination == suit and
                partners_last_bid.level == level):
            bid = Pass('0867')
        elif (suit.is_suit and
                self.next_level(suit) <= level):
            bid = self.suit_bid(level, suit, '0868')
        elif self.nt_level <= level:
            bid = self.nt_bid(level, '0869')
        else:
            bid = Pass('0870')
        if bid.name == self.bid_history[-2]:
            bid = Pass('0871')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def partnership_aces(self):
        """Return the number of aces in the partnership."""
        aces = self.aces
        if self.bid_history[-4] == '4NT':
            aces = (Bid(self.bid_history[-2]).denomination.rank + self.aces)
        elif self.bid_history[-8] == '4NT':
            aces = (Bid(self.bid_history[-6]).denomination.rank + self.aces)
        return aces

    def partnership_kings(self):
        """Return the number of kings in the partnership."""
        kings = self.kings
        if self.bid_history[-4] == '5NT':
            kings = (Bid(self.bid_history[-2]).denomination.rank + self.kings)
            if kings == 0:
                kings = 4
        return kings

    def select_contract_suit(self):
        """Return the selected suit after Blackwood process."""
        if (self.openers_first_bid.name == '2C' and
                self.opener_bid_two.name == '4NT' and
                self.suit_length(self.responder_bid_one.denomination) >= 3):
            suit = self.responder_bid_one.denomination
        elif (self.openers_first_bid.name == '2C' and
                self.suit_length(self.opener_bid_two.denomination) >= 3):
            suit = self.opener_bid_two.denomination
        elif self.responders_first_bid.denomination == self.openers_first_bid.denomination:
            suit = self.openers_first_bid.denomination
        elif self.responders_first_bid.denomination == self.openers_second_bid.denomination:
            suit = self.responders_first_bid.denomination
        elif self.responders_second_bid.denomination == self.openers_second_bid.denomination:
            suit = self.openers_second_bid.denomination
        elif self.responders_second_bid.denomination == self.openers_first_bid.denomination:
            suit = self.openers_first_bid.denomination
        elif self.responders_first_bid.denomination == self.opener_bid_three.denomination:
            suit = self.responders_first_bid.denomination
        elif (self.suit_length(self.openers_first_bid.denomination) >= 3 and
                len(self.bid_history) % 4 == 2):
            suit = self.openers_first_bid.denomination
        elif (self.is_jump(self.openers_first_bid, self.responders_first_bid) and
              self.hcp >= 16 and
              self.openers_first_bid.denomination.is_minor and
              self.responders_first_bid.denomination.is_minor):
            suit = self.no_trumps
        elif self.shape[0] >= 7:
            suit = self.longest_suit
        else:
            suit = self.no_trumps
        self.tracer(__name__, inspect.currentframe(), suit, self.trace)
        return suit



""" Bid for Game
    Acol OverCallersBid module
"""

import inspect
from .bidding import Pass, Double
from .hand import Hand


class OverCallersThirdBid(Hand):
    """BfG OverCallersRebid class."""
    def __init__(self, hand_cards, board):
        super(OverCallersThirdBid, self).__init__(hand_cards, board)
        self.trace = 0

    def suggested_bid(self):
        """Direct control to relevant method and return a Bid object."""
        if self._advancer_supports_major_after_stayman():
            if self.hcp >= 16:
                bid = self.next_level_bid(self.bid_two.denomination, '0842')
            else:
                bid = Pass('0843')
        elif self._can_show_second_major_after_stayman():
            bid = self.next_level_bid(self.spade_suit, '0216')
        elif self._has_thirteen_points_advancer_supports_second_suit():
            bid = self.next_level_bid(self.bid_two.denomination, '0844')
        elif self._has_fifteen_points_advancer_bids_two_suits():
            bid = self._fifteen_points_advancer_bids_two_suits()
        elif self._is_very_strong_and_five_four():
            if self.second_suit not in self.opponents_suits:
                suit = self.second_suit
            else:
                suit = self.longest_suit
            bid = self.next_level_bid(suit, '0846')
        elif self._is_strong_and_advancer_bids():
            bid = self._strong_hand()
        elif self._three_suits_have_been_bid():
            bid = self._three_suits_bid()
        elif self._has_two_card_support_for_advancers_major():
            bid = self.next_level_bid(self.partner_bid_one.denomination, '0324')
        elif self._has_six_card_suit_advancer_repeats_suit():
            bid = self.next_level_bid(self.longest_suit, '0850')
        elif self._is_strong_advancer_bids_nt_after_stayman():
            bid = self.nt_bid(3, '0000')
        else:
            bid = Pass('0851')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _strong_hand(self):
        """Return bid with strong hand."""
        if self.bid_one.is_double and self.bid_two.is_double:
            bid = Pass('0132')
        elif self._support_for_advancer():
            bid = self.next_level_bid(self.advancer_bid_one.denomination, '0658')
        elif self.is_balanced and self.nt_level <= 3:
            bid = self.next_nt_bid('0852')
        elif self.five_five and self.second_suit not in self.opponents_suits:
            bid = self.next_level_bid(self.second_suit, '0853')
        elif self.five_four and self.second_suit not in self.opponents_suits:
            bid = self.next_level_bid(self.second_suit, '0854')
        elif (self.partner_bid_one.denomination != self.partner_last_bid.denomination and
                self.longest_suit not in self.opponents_suits):
            bid = self.next_level_bid(self.longest_suit, '0855')
        elif (not self.partner_bid_one.is_pass and
                self.suit_length(self.partner_bid_one.denomination) >= 3):
            bid = self.next_level_bid(self.partner_bid_one.denomination, '0857')
        else:
            bid = Pass('0858')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _three_suits_bid(self):
        """Return bid after 3 suits have been bid."""
        suit_to_bid = self._select_suit_after_three_suits_bid()
        if self._suit_is_minor(suit_to_bid):
            bid = self.next_nt_bid('0847')
        elif suit_to_bid not in self.opponents_suits:
            bid = self.next_level_bid(suit_to_bid, '0848')
        else:
            bid = Pass('0849')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _fifteen_points_advancer_bids_two_suits(self):
        """Return bid if strong and advancer bids two suits."""
        if self.bidding_above_game:
            bid = Double('0912')
        elif self._advancer_has_passed_after_double():
            bid = Pass('0913')
        elif (self.suit_length(self.opener_bid_one.denomination) >= 3 and
                self.suit_points(self.opener_bid_one.denomination) >= 5):
            bid = Pass('0427')
        else:
            suit = self._suit_preference()
            bid = self.next_level_bid(suit, '0845')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    # Various utility functions

    def _suit_preference(self):
        """Return preference (if any) for partner's suits."""
        suit = None
        if (self.partner_bid_one.denomination.is_major or
                self.partner_bid_one.denomination.is_minor):
            suit_one = self.partner_bid_one.denomination
            if self.suit_length(suit_one) >= 3:
                suit = suit_one
        if (self.partner_last_bid.denomination.is_major or
                self.partner_last_bid.denomination.is_minor):
            suit_two = self.partner_last_bid.denomination
            if (self.suit_length(suit_two) >= 4 or
                    not suit):
                suit = suit_two
        if not suit:
            suit = self.no_trumps
        return suit

    def _select_suit_after_three_suits_bid(self):
        if self.suit_length(self.partner_last_bid.denomination) >= 4:
            suit_to_bid = self.partner_last_bid.denomination
        elif self.five_five:
            suit_to_bid = self.second_suit
            if suit_to_bid in self.opponents_suits:
                suit_to_bid = self.longest_suit
        else:
            suit_to_bid = self.longest_suit
        return suit_to_bid

    # Various boolean functions

    def _advancer_supports_major_after_stayman(self):
        """Return True if advancer supports major after Stayman."""
        result = (self.bid_one.name == '1NT' and
                  self.partner_bid_one.name == '2C' and
                  self.partner_last_bid.denomination == self.bid_two.denomination and
                  self.partner_last_bid.level == 3)
        return result

    def _can_show_second_major_after_stayman(self):
        """Return True if got 4 spades and 4 hearts."""
        result = (self.bid_one.name == '1NT' and
                  self.partner_bid_one.name == '2C' and
                  self.bid_two.name == '2H' and
                  self.partner_last_bid.name == '2NT' and
                  self.spades >= 4 and
                  self.spade_suit not in self.opponents_suits)
        return result

    def _has_thirteen_points_advancer_supports_second_suit(self):
        """Return True if 13 points and advancer support second suit."""
        result = (self.partner_last_bid.denomination == self.bid_two.denomination and
                  self.partner_last_bid.is_suit_call and
                  not self.partner_last_bid.is_game and
                  self.hcp >= 13 and
                  not self.singleton_honour)
        return result

    def _has_fifteen_points_advancer_bids_two_suits(self):
        """Return True if fifteen points and advancer bids two suits."""
        result = (self.partner_bid_one.denomination != self.partner_last_bid.denomination and
                  not self.bid_one.is_nt and
                  not self.partner_last_bid.is_nt and
                  not self.partner_last_bid.is_game and
                  self._suit_preference().is_suit and
                  self.hcp >= 15)
        return result

    def _is_very_strong_and_five_four(self):
        """Return True if strong and 5/4."""
        result = (self.hcp >= 19 and
                  self.five_four_or_better and
                  not self.bid_two.is_nt and
                  not self.bidding_above_game and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _is_strong_and_advancer_bids(self):
        """Return True if strong and advancer bids."""
        result = (self.hcp >= 16 and
                  not self.partner_last_bid.is_pass and
                  not self.partner_last_bid.is_game)
        return result

    def _three_suits_have_been_bid(self):
        """Return True if 3 suits bid."""
        my_suits = [self.bid_one.denomination, self.bid_two.denomination]
        result = (self.partner_bid_one.denomination != self.partner_last_bid.denomination and
                  self.partner_bid_one.is_suit_call and
                  self.partner_last_bid.level == 3 and
                  self.bid_one.denomination != self.partner_last_bid.denomination and
                  not self.partner_last_bid.is_game and
                  self.partner_last_bid.denomination not in my_suits)
        return result

    def _advancer_has_passed_after_double(self):
        """Return True if advancer passes after double."""
        result = (self.bid_one.is_double and
                  (self.partner_bid_one.is_pass or
                   self.partner_last_bid.is_pass))
        return result

    def _has_two_card_support_for_advancers_major(self):
        """Return True if 2 card support for advancers major."""
        result = (self.partner_bid_one.denomination == self.partner_last_bid.denomination and
                  self.partner_bid_one.is_value_call and
                  self.partner_bid_one.denomination.is_major and
                  self.suit_length(self.partner_bid_one.denomination) >= 2 and
                  self.hcp >= 15 and
                  not self.partner_last_bid.is_game)
        return result

    def _has_six_card_suit_advancer_repeats_suit(self):
        """Return True if 6 card suit and advancer repeats their suit."""
        result = (self.partner_bid_one.denomination == self.partner_last_bid.denomination and
                  self.partner_bid_one.is_value_call and
                  self.shape[0] >= 6 and
                  not self.partner_last_bid.is_game and
                  self.longest_suit not in self.opponents_suits and
                  self.hcp >= 12)
        return result

    def _is_strong_advancer_bids_nt_after_stayman(self):
        """Return True if strong and advancer bids NT after stayman."""
        result = (self.bid_one.name == '1NT' and
                  self.partner_bid_one.name == '2C' and
                  self.partner_last_bid.name == '2NT' and
                  self.hcp >= 17 and
                  self.nt_level <= 3)
        return result

    def _suit_is_minor(self, suit_to_bid):
        """Return True if suit is minor."""
        result = (self.nt_level <= 3 and
                  (suit_to_bid.is_minor or
                   suit_to_bid in self.opponents_suits))
        return result

    def _support_for_advancer(self):
        """Return True if xxx."""
        if not self.advancer_bid_two:
            result = False
        else:
            result = (self.advancer_bid_one.denomination == self.advancer_bid_two.denomination and
                      self.suit_length(self.advancer_bid_one.denomination) >= 3)
        return result

    # def xxx(self):
    #     """Return True if xxx."""
    #     result = False
    #     return result

    # def xxx(self):
    #     """Return True if xxx."""
    #     result = False
    #     return result



""" Bid for Game
    Acol OpenersReBid module
"""

import inspect
from .bidding import Bid, Pass, Double
from .hand import Hand
from .blackwood import Blackwood


class OpenersReBid(Hand):
    """BfG OpenersReBid class."""
    def __init__(self, hand_cards, board):
        super(OpenersReBid, self).__init__(hand_cards, board)
        self.trace = 0

    def suggested_bid(self):
        """Direct control to relevant method and return a Bid object."""
        if (self.responder_bid_one.name == '3NT' and
                self.shape[0] <= 5 and
                self.hcp <= 17):
            bid = Pass('0162')
        elif self.bid_one.is_nt:
            bid = self._rebid_after_nt_opening()
        elif self.responder_bid_one.is_pass:
            bid = self._responder_has_passed()
        elif self.bid_one.name == '2NT' and self.responder_bid_one.name == '4NT':
            if self.hcp == 22:
                bid = self.nt_bid(6, '0000')
            else:
                bid = Pass('0333')
        elif self.responder_bid_one.name == '4NT':
            bid = Blackwood(self.cards, self.board).count_aces()
        elif self.bid_one.name == '2C':
            bid = self._rebid_after_two_club_opening()
        elif (self.bid_one.level == 2 and
              self.bid_one.is_suit_call):
            bid = self._rebid_after_weak_two()
        elif (self.overcall_made and
              self.responder_bid_one.level == 3):
            bid = self._responder_at_three_overcall()
        elif self.responder_bid_one.is_double:
            bid = self._responder_has_doubled()
        else:
            bid = self._rebid_after_suit_opening()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _responder_has_passed(self):
        """Bid after responder has passed."""
        if (self.overcall_made and
                self.bid_one.level == 1):
            bid = self._responder_pass_overcall()
        elif self.bid_history[-3:] == ['P', 'P', 'P']:
            bid = Pass('0942')
        else:
            bid = Pass('0163')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _rebid_after_nt_opening(self):
        """Rebid after NT opening."""
        if self.bid_one.name == '1NT':
            bid = self._rebid_after_one_nt_opening()
        elif self.bid_one.name == '2NT':
            bid = self._rebid_after_two_nt_opening()
        elif self.bid_one.name == '3NT':
            bid = Pass('0000')
        else:
            assert False, 'Bid not defined'
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _rebid_after_one_nt_opening(self):
        """Rebid after 1NT opening."""
        if self.responder_bid_one.is_game:
            bid = Pass('0164')
        elif self.responder_bid_one.name == 'NT':
            bid = self._after_two_nt_response()
        elif self.responder_bid_one.name == '2NT':
            bid = self._after_two_nt_response()
        elif self.overcaller_bid_one.is_double:
            bid = self._one_nt_is_doubled()
        elif self.responder_bid_one.name == '2C':
            bid = self._after_stayman_two_clubs()
        elif self.responder_bid_one.level == 2:
            bid = self._after_one_nt_at_level_two()
        elif self.responder_bid_one.level == 3:
            bid = self._after_one_nt_at_level_three()
        else:
            bid = Pass('0588')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _after_one_nt_at_level_two(self):
        """Rebid after 1NT opening, responder at level 2."""
        if self._three_card_support_for_responder_overcall_made():
            bid = self.next_level_bid(self.responder_bid_one.denomination, '0166')
        elif self._four_card_support_for_responder_overcall_made():
            bid = self.next_level_bid(self.responder_bid_one.denomination, '0167')
        else:
            bid = Pass('0168')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _after_one_nt_at_level_three(self):
        """Rebid after 1NT opening, responder at level 3."""
        if (self.responder_bid_one.name == '3H' or
                self.responder_bid_one.name == '3S'):
            bid = self._after_three_of_major_response()
        elif (self.five_card_major and
              self.longest_suit not in self.opponents_suits):
            bid = self.next_level_bid(self.longest_suit, '0169')
        elif self.nt_level <= 3 and self.stoppers_in_bid_suits:
            bid = self.nt_bid(3, '0170')
        elif self.longest_suit not in self.opponents_suits:
            bid = self.next_level_bid(self.longest_suit, '0807')
        else:
            bid = Pass('0172')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _one_nt_is_doubled(self):
        """Rebid after 1NT has been doubled."""
        if Bid(self.bid_history[-1]).is_value_call:
            bid = Pass('0172')
        elif self.responder_bid_one.is_value_call:
            bid = Pass('0579')
        else:
            bid = Pass('0585')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _after_stayman_two_clubs(self):
        """Rebid after Stayman 2C."""
        if self.overcaller_bid_one.is_double:
            bid = Pass('0172')
        elif self.nt_level >= 3:
            bid = Pass('0574')
        elif (self.hearts >= 4 and
              self.next_level(self.heart_suit) <= 2):
            bid = self.heart_bid(2, '0173')
        elif (self.spades >= 4 and
              self.next_level(self.spade_suit) <= 2):
            bid = self.spade_bid(2, '0174')
        elif self.next_level(self.diamond_suit) <= 2:
            bid = self.diamond_bid(2, '0175')
        elif self.opener_bid_one.name != '1NT' or self.stoppers_in_bid_suits:
            bid = self.nt_bid(2, '0176')
        else:
            bid = Pass('0130')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _after_stayman_three_clubs(self):
        """Rebid after Stayman."""
        if self.overcaller_bid_one.is_double:
            bid = Pass('0854')
        elif (self.overcall_made and
                self.nt_level >= 3):
            bid = Pass('0855')
        elif (self.hearts >= 4 and
              self.next_level(self.heart_suit) <= 3):
            bid = self.heart_bid(3, '0856')
        elif (self.spades >= 4 and
              self.next_level(self.spade_suit) <= 3):
            bid = self.spade_bid(3, '0857')
        elif self.next_level(self.diamond_suit) <= 3:
            bid = self.diamond_bid(3, '0858')
        else:
            bid = self.nt_bid(3, '0859')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _rebid_after_two_nt_opening(self):
        """Rebid after 2NT opening."""
        if self.responder_bid_one.name == '3NT':
            bid = Pass('0177')
        elif self.bid_one.name == '2NT' and self.responder_bid_one.name == '4NT':
            if self.hcp == 22:
                bid = self.nt_bid(6, '0707')
            else:
                bid = Pass('0333')
        elif self.responder_bid_one.name == '3C':
            bid = self._after_stayman_three_clubs()
        elif (self.suit_length(self.responder_bid_one.denomination) >= 3 and
                not self.bidding_above_game):
            bid = self.bid_to_game(self.responder_bid_one.denomination, '0875')
        elif (self.responder_bid_one.is_suit_call and
                self.nt_level <= 3):
            bid = self.nt_bid(3, '0876')
        elif self.responder_bid_one.name == '4NT' and self.hcp <= 22:
            bid = Pass('0333')
        else:
            bid = Pass('0178')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _responder_pass_overcall(self):
        """Rebid after an overcall and responder has passed."""
        first_suit, second_suit = self._get_best_suits()
        level = self.next_level(first_suit)
        if self.hcp >= 19:
            bid = self._respond_with_power(first_suit, second_suit)
        elif self._is_strong_with_five_five(second_suit):
            bid = self.next_level_bid(second_suit, '0179')
        elif self._strong_with_second_suit(level):
            bid = self.next_level_bid(second_suit, '0180')
        elif self._six_card_suit_weak():
            bid = self.next_level_bid(first_suit, '0181')
        elif self._is_strong_with_five_four(second_suit):
            bid = self.next_level_bid(second_suit, '0695')
        elif self._repeatable_five_card_suit():
            if self.hcp >= 18 and self.shape[0] >= 6 and self.overcaller_in_second_seat:
                raise_level = 1
            else:
                raise_level = 0
            bid = self.next_level_bid(first_suit, '0182', raise_level=raise_level)
        elif self._strong_and_balanced():
            bid = self.next_nt_bid('0183')
        elif self.hcp >= 16 and self.shape[0] >= 6:
            suit = self._select_suit_if_six_four()
            bid = self.next_level_bid(suit, '0184')
        else:
            bid = self._responder_pass_overcall_weak()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _responder_pass_overcall_weak(self):
        """Rebid after an overcall and responder has passed weakish hand."""
        first_suit = self.longest_suit
        if self._weak_five_four():
            bid = self._weak_five_four_bid()
        elif self._weak_six_cards():
            bid = self.next_level_bid(first_suit, '0185')
        elif self.five_five and self.nt_level <= 3:
            bid = self._five_five_rebid()
        elif self._can_rebid_five_card_suit_at_two_level():
            bid = self.next_level_bid(self.longest_suit, '0186')
        elif self._can_rebid_seven_card_suit_at_three_level():
            bid = self.next_level_bid(self.longest_suit, '0878')
        else:
            bid = Pass('0187')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _weak_five_four_bid(self):
        """Return bid if weak but biddable 5/4."""
        barrier_will_be_broken = self.barrier_is_broken(self.bid_one,
                                                        self.next_level_bid(self.second_suit))
        will_be_jump_bid = self.is_jump(self.bid_one,
                                        self.next_level_bid(self.longest_suit))
        if will_be_jump_bid:
            bid = Pass('0188')
        elif barrier_will_be_broken:
            bid = self.next_level_bid(self.longest_suit, '0189')
        else:
            bid = self.next_level_bid(self.second_suit, '0190')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _five_five_rebid(self):
        """Bid with 5/5 hand"""
        suit = self._other_five_card_suit()
        if self._overcaller_doubled_and_advancer_bid():
            bid = Pass('0191')
        elif self.next_level(suit) > 3:
            bid = Pass('0192')
        elif suit == self.bid_one.denomination and self.hcp <= 13:
            bid = Pass('0906')
        elif suit not in self.opponents_suits and self.hcp >= 12:
            bid = self.next_level_bid(suit, '0193')
        elif self.bid_one.denomination not in self.opponents_suits and self.hcp >= 12:
            bid = self.next_level_bid(self.bid_one.denomination, '0194')
        else:
            bid = Pass('0195')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _respond_with_power(self, first_suit, second_suit):
        """Rebid after an overcall and responder has passed 19+ points."""
        if self.six_four:
            bid = self._powerful_six_four_bid(first_suit, second_suit)
        elif self.five_four_or_better:
            if second_suit not in self.opponents_suits:
                suit = second_suit
            else:
                suit = first_suit
            bid = self._powerful_five_four_bid(suit)
        elif self.stoppers_in_bid_suits and self.nt_level <= 3:
            bid = self.next_nt_bid('0196')
        else:
            bid = Double('0197')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _powerful_six_four_bid(self, first_suit, second_suit):
        """Bid powerful hand with 6/4."""
        first_suit_points = self.suit_points(first_suit)
        second_suit_points = self.suit_points(second_suit)
        if first_suit_points > second_suit_points+1:
            suit = first_suit
        else:
            suit = second_suit
        if suit in self.opponents_suits:
            if suit == first_suit:
                suit = second_suit
            else:
                suit = first_suit
        if (self.next_level(suit) <= 2 and
                (suit == self.opener_bid_one.denomination or
                 suit.rank < self.opener_bid_one.denomination.rank)):
            raise_level = 1
        else:
            raise_level = 0
        bid = self.next_level_bid(suit, '0198', raise_level)
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _powerful_five_four_bid(self, second_suit):
        """Bid powerful hand with 5/4."""
        suit = second_suit
        raise_level = 0
        trial_bid = self.next_level_bid(suit, '0199', raise_level)
        level = trial_bid.level
        if ((trial_bid.level <= 3 and
             self.responder_bid_one.is_value_call and
             not self.overcall_made) or
                self.hcp >= 22):
            level = 3
        next_level = self.next_level(suit)
        level = max(next_level, level)
        bid = self.suit_bid(level, suit, '0200')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _rebid_after_two_club_opening(self):
        """Bid after a two club opening."""
        if self.is_balanced:
            bid = self._balanced_after_two_clubs()
        else:
            bid = self._rebid_no_support()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _balanced_after_two_clubs(self):
        """Bid with a balanced hand after two club opening."""
        if self.responder_bid_one.name == '2NT':
            bid = self.nt_bid(3, '0600')
        else:
            bid = self._balanced_two_clubs_responder_with_values()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _balanced_two_clubs_responder_with_values(self):
        """Bid with a balanced hand after two club opening with support."""
        if self.hcp >= 28:
            bid = self.nt_bid(4, '0202')
        elif self.responder_bid_one.is_major and self.suit_length(self.responder_bid_one.denomination):
            bid = self.next_level_bid(self.responder_bid_one.denomination, '0941')
        elif ((self.hcp >= 25 or self.responder_bid_one.level == 3) and
              self.nt_level <= 3):
            bid = self.nt_bid(3, '0203')
        elif self.nt_level <= 2:
            bid = self.nt_bid(2, '0204')
        else:
            bid = self.next_level_bid(self.responder_bid_one.denomination, '0000')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _rebid_after_weak_two(self):
        """Rebid after a weak two opening."""
        if self.responder_bid_one.denomination == self.bid_one.denomination:
            bid = Pass('0205')
        else:
            bid = self._rebid_after_suit_opening()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _after_two_nt_response(self):
        """Rebid after 2NT response from responder."""
        if self.five_card_major and self.longest_suit not in self.opponents_suits:
            bid = self.next_level_bid(self.longest_suit, '0687')
        elif self.nt_level <= 3:
            bid = self._after_two_nt_response_bid_nt()
        else:
            bid = Pass('0206')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _after_two_nt_response_bid_nt(self):
        """Rebid after 2NT response from responder, bid NT."""
        if (self.hcp == 13 and (self.shape[0] == 5 or
                                self.tens_and_nines >= 4)):
            bid = self.nt_bid(3, '0207')
        elif self.hcp == 14:
            bid = self.nt_bid(3, '0208')
        else:
            bid = Pass('0209')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _after_three_of_major_response(self):
        """Rebid after 3 major response."""
        suit = self.responder_bid_one.denomination
        level = self.next_level(suit)
        if self.responder_bid_one.name == '3H' and self.hearts >= 3 and level <= 4:
            bid = self.heart_bid(4, '0210')
        elif self._can_rebid_spades(level):
            bid = self.spade_bid(4, '0211')
        elif self.nt_level <= 3:
            bid = self.nt_bid(3, '0212')
        else:
            bid = Pass('0213')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _rebid_after_suit_opening(self):
        """Rebid after suit opening."""
        if self.responder_bid_one.is_nt:
            bid = self._rebid_after_nt_response()
        elif self._responder_shows_support_or_can_support_responder():
            bid = self._after_suit_opening_support()
        elif self._can_support_responders_major():
            if self.hcp <= 12:
                raise_level = 0
            else:
                raise_level = 1
            bid = self.next_level_bid(self.responder_bid_one.denomination, '0214', raise_level)
        else:
            bid = self._rebid_no_support()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _after_suit_opening_support(self):
        """Rebid after suit opening with support."""
        if self.responder_bid_one.is_minor and self.is_balanced:
            bid = self._rebid_no_support()
        else:
            bid = self._rebid_with_support()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _rebid_after_nt_response(self):
        """Rebid after NT response."""
        if self._responder_has_bid_nt_game_over_minor():
            bid = Pass('0215')
        elif self._is_shapely_intermediate_hand():
            bid = self._rebid_no_support()
        else:
            bid = self._support_nt()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _support_nt(self):
        """Support NT bid from partner."""
        if self.responder_bid_one.name == '1NT':
            bid = self._support_one_nt()
        elif self.responder_bid_one.name == '2NT':
            bid = self._support_two_nt()
        elif self.responder_bid_one.name == '3NT':
            bid = Pass('0216')
        else:
            bid = Pass('999')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _support_one_nt(self):
        """Respond to 1NT bid from partner."""
        old_level = self.responder_bid_one.level
        level = self.quantitative_raise(self.hcp, old_level, [16, 19], 3)
        if self.hcp >= 16 and self.nt_level <= level:
            bid = self.nt_bid(level, '0217')
        elif self.shape[0] >= 5 and (self.last_bid.is_pass or self.shape[0] >= 6):
            bid = self._rebid_no_support()
        elif self.hcp >= 13 and self.overcall_made:
            bid = self._rebid_no_support()
        elif self.is_balanced:
            bid = Pass('0218')
        else:
            bid = Pass('0879')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _support_two_nt(self):
        """Support 2NT bid from partner."""
        if self.hcp >= 15 and self.nt_level <= 3:
            bid = self.nt_bid(3, '0219')
        else:
            bid = Pass('0220')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _rebid_with_support(self):
        """Rebid after suit decided."""
        if self.bid_one.is_suit_call:
            change_suit = (self.responder_bid_one.denomination != self.bid_one.denomination and
                           self.responder_bid_one.denomination.is_suit and
                           self.bid_one.denomination.is_suit)
        else:
            change_suit = None
        if change_suit:
            bid = self._opener_can_support_responder()
        else:
            bid = self._rebid_support_no_suit_changes()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _rebid_support_no_suit_changes(self):
        """Rebid responder supports opener."""
        if self.responder_bid_one.level == 4:
            if (self.hcp >= 19 or
                (self.hcp >= 17 and
                 self.bid_one.denomination.is_major)):
                if self.nt_level <= 4:
                    bid = self.nt_bid(4, '0317')
                else:
                    bid = self.next_level_bid(self.responder_bid_one.denomination, '0000')
            elif self.hcp >= 16:
                bid = self.next_level_bid(self.responder_bid_one.denomination, '0925')
            else:
                bid = Pass('0221')
        elif self.responder_bid_one.level >= 3:
            bid = self._support_at_level_three()
        elif self.responder_bid_one.level >= 2:
            bid = self._support_at_level_two()
        else:
            bid = Pass('0222')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _support_at_level_three(self):
        """Rebid with support and no suit changes at level 3."""
        next_level = self.next_level(self.responder_bid_one.denomination)
        if (self.responder_bid_one.denomination.is_major and self.hcp >= 14 and
                next_level <= 4):
            bid = self.bid_to_game(self.responder_bid_one.denomination, '0223')
        elif (self.responder_bid_one.denomination.is_minor and self.hcp >= 16 and
              next_level <= 5):
            bid = self.bid_to_game(self.responder_bid_one.denomination, '0224')
        else:
            bid = Pass('0225')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _support_at_level_two(self):
        """Rebid with support and no suit changes at level 2."""
        agreed_suit = self.my_last_bid.denomination
        next_level = self.next_level(agreed_suit)
        game_bid = self.bid_to_game(agreed_suit)
        if self.hcp >= 19 and next_level <= game_bid.level:
            bid = self.bid_to_game(agreed_suit, '0226')
        elif self._has_six_card_minor_and_eighteen_points():
            bid = self.nt_bid(3, '0227')
        elif self.hcp >= 16 and next_level <= 4:
            bid = self.next_level_bid(agreed_suit, '0228')
        elif self.overcall_made and self.hcp >= 15 and next_level <= 3:
            bid = self.next_level_bid(agreed_suit, '0229')
        elif self.shape[0] >= 7:
            bid = self.next_level_bid(agreed_suit, '0230')
        elif self.hcp >= 14 and self.shape[0] >= 5 and not self.right_hand_bid.is_pass:
            bid = self.next_level_bid(agreed_suit, '0756')
        elif self.shape[0] >= 6 and self.shape[1] >= 5:
            bid = self.next_level_bid(self.longest_suit, '0792')
        else:
            bid = Pass('0231')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _opener_can_support_responder(self):
        """Return bid when opener supports responder."""
        suit_to_bid = self._get_suit_support_for_responder()
        level = self._get_level_with_support(suit_to_bid)
        next_level = self.next_level_bid(suit_to_bid).level
        level = max(level, next_level)
        responder_has_jumped = (self.is_jump(self.bid_one, self.responder_bid_one))

        if self.hcp >= 16 and responder_has_jumped and self.overcaller_bid_one.is_pass:
            bid = self._opener_can_support_responder_strong(suit_to_bid)
        elif suit_to_bid.is_no_trumps:
            bid = Pass('0234')
        elif self.suit_length(suit_to_bid) <= 3:
            if self.responder_bid_one.level == 2 and self.hcp >= 14:
                bid = self.next_level_bid(suit_to_bid, '0263', raise_level=1)
            else:
                bid = self.next_level_bid(suit_to_bid, '0235')
        elif self._is_unbalanced_with_sixteen_points():
            bid = self._medium_unbalanced_hand()
        elif level >= next_level:
            bid = self.suit_bid(level, suit_to_bid, '0236', use_shortage_points=True)
        else:
            bid = Pass('0237')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _medium_unbalanced_hand(self):
        """Return bid with 16+ points and unbalanced."""
        if self.overcall_made:
            raise_level = 0
        else:
            raise_level = 1
        if self._can_support_nineteen_points():
            bid = self.bid_to_game(self.second_suit, '0883')
        elif self._can_support_seventeen_points_and_void():
            bid = self.bid_to_game(self.responder_bid_one.denomination, '0733')
        elif self._fffo_strong_four_card_support():
            bid = self.bid_to_game(self.responder_bid_one.denomination, '0327')
        elif self._no_biddable_second_suit_and_support_for_partner():
            bid = self.bid_to_game(self.responder_bid_one.denomination, '0689')
        elif self.second_suit == self.responder_bid_one.denomination:
            bid = self.next_level_bid(self.second_suit, '0203', raise_level=raise_level)
        else:
            bid = self.next_level_bid(self.second_suit, '0165', raise_level=raise_level)
            # if bid.level > self.game_level(self.second_suit):
            #     bid = Double('0203')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _opener_can_support_responder_strong(self, suit_to_bid):
        """Return bid when opener supports responder and 16+ points."""
        if self.hcp >= 16 and suit_to_bid.is_major:
            if self.responder_bid_one.level == 3:
                bid = self.nt_bid(4, '0232')
            else:
                bid = self.nt_bid(2, '0264')
        else:
            bid = self.next_level_bid(suit_to_bid, '0233')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _rebid_no_support(self):
        """Rebid with no responder support."""
        trial_bid = self._five_four_rebids()
        if self.bid_one.name == '2C':
            bid = trial_bid
        elif self.is_balanced:
            bid = self._no_support_balanced()
        elif self._has_strong_minor_responder_has_jumped():
            bid = self.nt_bid(4, '0239')
        elif self.shape[0] >= 7:
            bid = self._single_suited_rebids()
        elif self.five_five:
            bid = self._no_support_five_five()
        elif self._competitive_auction_weak_with_minor() and trial_bid.level >= 3:
            bid = Pass('0240')
        elif self._competitive_auction_with_support_for_responder():
            bid = self._rebid_partners_major()
        elif self.five_in_opponents_suits():
            bid = Pass('0241')
        elif self.five_four or self.five_five_or_better:
            bid = self._five_four_rebids()
        elif self.shape[0] >= 6:
            bid = self._single_suited_rebids()
        elif self.shape == [4, 4, 4, 1]:
            bid = self._four_four_four_one_bid()
        else:
            bid = Pass('0000')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _no_support_five_five(self):
        """Rebid with no responder support and 5/5 hand."""
        suit_to_bid = self._get_suit_no_support_five_five()
        raise_level = 1
        level = min(self.next_level(suit_to_bid, raise_level=raise_level), 3)

        if self._suit_is_major_and_eighteen_points(suit_to_bid):
            if self.responder_bid_one.is_nt:
                bid = self.suit_bid(4, suit_to_bid, '0880')
            else:
                bid = self.suit_bid(level, suit_to_bid, '0201')
        elif self.next_level(suit_to_bid) <= 3 and self.hcp >= 16:
            bid = self.suit_bid(level, suit_to_bid, '0242')
        elif self.five_in_opponents_suits():
            bid = Pass('0243')
        elif (self.responder_bid_one.name != '2NT' and
                (self.next_level(suit_to_bid) <= 2 or
                 self.hcp >= 15 or
                 self.overcaller_bid_one.denomination == self.advancer_bid_one.denomination)):
            bid = self.next_level_bid(suit_to_bid, '0244')
        else:
            bid = Pass('0335')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _rebid_partners_major(self):
        """Rebid with no responder support competitive auction and partner major."""
        responders_suit = self.responder_bid_one.denomination
        if (self.hcp >= 14 and
                self.support_shape_points(responders_suit) >= 2):
            bid = self.next_level_bid(responders_suit, '0245', raise_level=1)
        else:
            bid = self.next_level_bid(responders_suit, '0246')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _no_support_balanced(self):
        """Bid with no support and balanced hand."""
        level = self.next_level(self.bid_one.denomination)
        if self.hcp >= 19:
            bid = self._bid_with_19_points()
        elif self.hcp >= 15 or level > 2:
            bid = self._invitational_bid()
        else:
            bid = self.next_level_bid(self.bid_one.denomination, '0247')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _bid_with_19_points(self):
        """Return bid with a 19 point hand."""
        unbid_major = self.unbid_four_card_major()
        if self._powerful_and_has_four_cards_in_unbid_major(unbid_major):
            bid = self.next_level_bid(unbid_major, '0248', raise_level=1)
        elif self._responder_support_with_jump():
            bid = self.nt_bid(3, '0943')
        elif (self.hcp >= 17 and self.is_jump(self.bid_one,
                                              self.responder_bid_one)):
            bid = self.nt_bid(6, '0015')
        elif self.nt_level <= 3:
            bid = self.nt_bid(3, '0250')
        else:
            bid = Pass('0251')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _invitational_bid(self):
        """Return bid with 15+ points or partner has 10+ points."""
        if self.stoppers_in_bid_suits:
            bid = self._rebid_balanced_hand()
        elif self.shape[1] >= 4 and self.hcp >= 16 and self.second_suit not in self.opponents_suits:
            bid = self.next_level_bid(self.second_suit, '0757')
        elif self.shape[0] < 5:
            bid = self._opener_can_support_responder()
        else:
            bid = self._single_suited_rebids()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _five_four_rebids(self):
        """Rebid with 5/4 hands."""
        responder_jumped = self.jump_bid_made(self.responder_bid_one)
        suit_to_bid = self._get_suit_with_five_four()
        test_bid = self.next_level_bid(suit_to_bid)
        cannot_show_second_suit = self.barrier_is_broken(self.bid_one, test_bid)
        if self._weak_and_able_to_repeat_suit(cannot_show_second_suit, responder_jumped):
            suit_to_bid = self.bid_one.denomination

        if self.hcp >= 16:
            bid = self._five_four_strong(suit_to_bid)
        else:
            bid = self._five_four_weak(suit_to_bid, cannot_show_second_suit)
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _five_four_strong(self, suit_to_bid):
        """Bid with strong hand."""
        if self.bid_one.name == '2C':
            bid = self._rebid_after_two_clubs()
        elif self.five_five:
            suit_to_bid = self.longest_suit
            if self.hcp >= 16:
                raise_level = 1
            else:
                raise_level = 0
            bid = self.next_level_bid(suit_to_bid, '0252', raise_level)
        else:
            bid = self._five_four_strong_balanced(suit_to_bid)
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _five_four_strong_balanced(self, suit_to_bid):
        """Bid with strong 5/4 hand."""
        test_bid = self.next_level_bid(suit_to_bid, '0253')
        if self._five_four_strong_test_nt(test_bid):
            bid = self.nt_bid(3, '0254')
        elif self._can_repeat_suit_and_seventeen_points(suit_to_bid):
            bid = self._five_four_strong_nt()
        elif self._has_five_four_and_sixteen_points(suit_to_bid):
            second_suit = self._get_second_suit_for_opener()
            bid = self.next_level_bid(second_suit, '0255')
        else:
            raise_level = self._raise_level_with_five_four_strong(suit_to_bid, test_bid)
            if (self.next_level(suit_to_bid, raise_level) > 3 or
                    (self.responder_bid_one.name == '1NT' and self.hcp <= 16)):
                raise_level = 0
            bid = self.next_level_bid(suit_to_bid, '0257', raise_level)
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _five_four_strong_nt(self):
        """Bid with strong 5/4 hand try NT."""
        if self.hcp >= 19 and self.nt_level <= 3:
            bid = self.nt_bid(3, '0258')
        elif self._can_bid_three_nt():
            bid = self.nt_bid(3, '0000')
        elif self._can_bid_two_nt():
            bid = self.nt_bid(2, '0259')
        elif self.nt_level <= 1:
            bid = self.nt_bid(1, '0260')
        else:
            if self.second_suit in self.opponents_suits:
                suit_to_bid = self.longest_suit
            else:
                suit_to_bid = self.second_suit
            bid = self.next_level_bid(suit_to_bid, '0261')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _rebid_after_two_clubs(self):
        """Return bid after a 2C opening."""
        if self.is_semi_balanced and self.hcp >= 25 and self.nt_level <= 3:
            bid = self.nt_bid(3, '0523')
        elif self.is_balanced:
            bid = self.next_nt_bid('0262')
        elif self.responder_bid_one.name == '2D' and self.longest_suit == self.diamond_suit:
            bid = Pass('0000')
        elif self.responder_bid_one.name == '2D':
            bid = self.next_level_bid(self.longest_suit, '0848')
        elif self.suit_length(self.responder_bid_one.denomination) >= 3:
            if (self.hcp + self.suit_length(self.responder_bid_one.denomination) >= 26 and
                    self.nt_level <= 4):
                bid = self.nt_bid(4, '0000')
            else:
                bid = self.bid_to_game(self.responder_bid_one.denomination, '0264')
        elif self.longest_suit.is_minor and self.responder_bid_one.denomination.is_minor:
            bid = self.next_nt_bid('0265')
        else:
            bid = self.next_level_bid(self.longest_suit, '0266')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _five_four_weak(self, suit, cannot_show_second_suit):
        """Bid with weak 5/4 hand."""
        suit_to_bid = self._five_four_weak_suit(suit, cannot_show_second_suit)
        level = self.next_level(suit)
        if self._weak_five_four_with_weak_suit(suit_to_bid):
            bid = Pass('0267')
        elif self._weak_five_four_consider_bid(level):
            bid = self._five_four_weak_bid(suit_to_bid, cannot_show_second_suit)
        elif self._can_bid_at_appropriate_level(suit_to_bid, level):
            bid = self.next_level_bid(suit_to_bid, '0268')
        elif self.bid_history[-1] == 'P':
            bid = self.next_level_bid(self.longest_suit, '0271')
        else:
            bid = Pass('0269')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _five_four_weak_bid(self, suit_to_bid, cannot_show_second_suit):
        """Return bid if a sound bid is possible."""
        overcall = Bid(self.bid_history[-1])
        trial_bid = self.next_level_bid(suit_to_bid, '0270')
        if trial_bid.level >= 3 or (not overcall.is_value_call):
            bid = self._five_four_weak_invitational(suit_to_bid, cannot_show_second_suit)
        elif self.hcp <= 11:
            bid = Pass('0271')
        elif self.hcp <= 12 and cannot_show_second_suit:
            bid = Pass('0272')
        else:
            bid = self.next_level_bid(suit_to_bid, '0273')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _five_four_weak_invitational(self, suit_to_bid, cannot_show_second_suit):
        """Return bid with 5/4 hand and next level is 3."""
        responder_jumped = self.is_jump(self.bid_one, self.responder_bid_one)
        if self.bid_one.level == 2:
            bid = Pass('0274')
        elif self.next_level(suit_to_bid) <= 2 and self.shape[0] >= 6:
            if self.shape[0] >= 7 and self.hcp >= 13:
                raise_level = 1
            else:
                raise_level = 0
            bid = self.next_level_bid(suit_to_bid, '0275', raise_level=raise_level)
        elif self._weak_after_one_nt_and_barrier_broken(cannot_show_second_suit):
            bid = Pass('0276')
        elif self._six_four_responder_at_level_two():
            bid = self.next_level_bid(self.second_suit, '0277')
        elif self._weak_overcall_responder_level_one() and cannot_show_second_suit:
            bid = Pass('0278')
        elif self._can_bid_second_suit_after_nt():
            bid = self.next_level_bid(self.second_suit, '0834')
        elif self.hcp <= 12 and cannot_show_second_suit:
            bid = self.next_level_bid(self.bid_one.denomination, '0873')
        elif self._strong_or_can_bid_at_level_two(suit_to_bid):
            bid = self.next_level_bid(suit_to_bid, '0279')
        elif responder_jumped and not self.responder_bid_one.is_nt:
            if self.second_suit not in self.opponents_suits:
                bid = self.next_level_bid(self.second_suit, '0000')
            else:
                bid = self.next_level_bid(self.longest_suit, '0000')
        elif self._strong_five_four_or_better():
            bid = self.next_level_bid(self.second_suit, '0130')
        else:
            bid = Pass('0908')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _single_suited_rebids(self):
        """Return bid with single suited hands."""
        if self.hcp >= 18 or self.shape[0] >= 8:
            bid = self._single_suited_rebids_strong()
        elif self.hcp >= 16:
            bid = self._single_suited_intermediate()
        else:
            bid = self._single_suited_rebids_weak()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _single_suited_rebids_strong(self):
        """Return bid with strong single suited hands."""
        suit = self.longest_suit
        level = self.next_level(suit)
        if suit.is_major:
            bid = self._single_suited_strong_major(suit, level)
        else:
            level = self.next_level(suit)
            level = max(3, level)
            bid = self.suit_bid(level, suit, '0280')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _single_suited_strong_major(self, suit, level):
        """Return bid with strong single major suited hands."""
        level = max(4, level)
        if (self.shape[0] <= 5 and
                self.nt_level <= 3):
            bid = self.nt_bid(3, '0281')
        elif level <= 5:
            bid = self.suit_bid(level, suit, '0282')
        else:
            bid = Pass('0283')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _single_suited_intermediate(self):
        """Return bid with intermediate single suited hands."""
        if self.hcp == 19:
            nt_level = 3
        elif self.hcp >= 17 or self.nt_level == 2:
            nt_level = 2
        else:
            nt_level = 1
        if self._is_strong_with_seven_card_major():
            bid = self.bid_to_game(self.longest_suit, '0903')
        elif self._is_strong_six_four_with_major():
            bid = self.next_level_bid(self.second_suit, '0256')
        elif self.shape[0] <= 5 and self.longest_suit.is_minor and self.nt_level <= nt_level:
            bid = self.nt_bid(nt_level, '0117')
        else:
            if self._weak_with_overcall_or_next_level_three():
                raise_level = 0
            else:
                raise_level = 1
            bid = self.next_level_bid(self.longest_suit, '0284', raise_level)
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _single_suited_rebids_weak(self):
        """Return bid with weak single suited hands."""
        if ((self.overcall_made == self.OVERCALLER_POSITION['fourth_seat']) and
                not self.responders_support):
            bid = self._single_suited_weak_competitive()
        else:
            bid = self._single_suited_uncontested()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _single_suited_uncontested(self):
        """Return bid with weak single suited hands in uncontested auction."""
        if self.responder_bid_one.is_game:
            bid = Pass('0285')
        elif self._has_minor_can_support_responders_major():
            bid = self.next_level_bid(self.responder_bid_one.denomination, '0286')
        elif self._has_six_card_major_and_responder_bid_one_nt():
            bid = self.suit_bid(2, self.longest_suit, '0912')
        elif self._has_six_card_major_and_responder_bid_two_nt():
            if ((self.shape[0] >= 7 and self.next_level(self.longest_suit) <= 3) or
                    (self.shape[0] == 6 and self.hcp >= 14 and
                     self.next_level(self.longest_suit) <= 4)):
                level = 4
            else:
                level = self.next_level(self.longest_suit)
            bid = self.suit_bid(level, self.longest_suit, '0287')
        elif self.responder_bid_one.name == '1NT' and self.hcp <= 15 and self.shape[0] <= 5:
            bid = Pass('0521')
        elif self.shape[0] >= 6 and self.longest_suit.is_major and self.hcp >= 12:
            if self.shape[0] >= 7 and self.hcp >= 13 and self.next_level(self.longest_suit) <= 2:
                raise_level = 1
            else:
                raise_level = 0
            bid = self.next_level_bid(self.longest_suit, '0929', raise_level)
        elif self._can_bid_three_nt():
            bid = self.next_nt_bid('0288')
        elif self.responder_bid_one.level >= 3 and self.suit_length(self.responder_bid_one.denomination) >= 3:
            bid = self.next_level_bid(self.responder_bid_one.denomination, '0360')
        elif self.hcp <= 9:
            bid = Pass('0836')
        else:
            bid = self.next_level_bid(self.longest_suit, '0890')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _single_suited_weak_competitive(self):
        """Return bid with weak single suited hands in competitive auction."""
        if self._is_strong_with_solid_seven_card_suit():
            bid = self.next_level_bid(self.longest_suit, '0891')
        elif self._has_five_card_suit_fifteen_points_level_two():
            bid = self.next_level_bid(self.longest_suit, '0033')
        elif self._has_seven_card_suit_and_weak():
            bid = self.next_level_bid(self.longest_suit, '0887')
        elif self.shape[0] >= 6 and self.next_level(self.longest_suit) <= 2:
            bid = self.next_level_bid(self.longest_suit, '0892')
        elif self._has_six_card_suit_responder_at_level_two():
            bid = self.next_level_bid(self.longest_suit, '0893')
        else:
            bid = Pass('0289')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _four_four_four_one_bid(self):
        """Rebid with 4441 hands."""
        singleton_suit = self.suit_shape[3]
        if self.responder_bid_one.denomination == singleton_suit:
            bid = self._fffo_partner_bids_singleton()
        else:
            bid = self._rebid_with_support()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _fffo_partner_bids_singleton(self):
        """Rebid with 4441 hands after partner bids singleton suit."""
        suit = self._select_suit_for_four_four_four_one()
        if self.responder_bid_one.level == 1:
            level = self.quantitative_raise(self.hcp, 0, [15, 17, 19], 3)
        else:
            level = self.quantitative_raise(self.hcp, 0, [12, 13, 15], 3)
        test_bid = self.next_level_bid(suit, '0290', False)
        if self._has_fifteen_points_and_level_is_one(level):
            bid = self.nt_bid(level, '0291')
        elif self._barrier_would_break_and_fewer_that_sixteen(suit, test_bid):
            bid = Pass('0292')
        else:
            bid = test_bid
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _rebid_balanced_hand(self):
        """Return bid with a balanced hand."""
        bid_level = self._balanced_bid_level()
        if self._strong_can_bid_nt(bid_level):
            comment = ['0067', '0293', '0065', '0066'][bid_level]
            bid = self.nt_bid(bid_level, comment)
        elif self._no_fit_fifteen_points_responder_level_two():
            bid = self.nt_bid(3, '0294')
        elif self.hcp >= 15 and self.is_balanced and self.stoppers_in_unbid_suits() and self.nt_level == 1:
            bid = self.nt_bid(1, '0816')
        else:
            bid = Pass('0295')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _responder_at_three_overcall(self):
        """Return bid after responder jumped to 3 level after an overcall."""
        if self.responder_bid_one.name == '3NT':
            bid = self._responder_at_three_nt_overcall()
        else:
            bid = self._at_three_overcall_suit()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _at_three_overcall_suit(self):
        """Return bid after responder jumped to 3 of a suit after overcall."""
        if self.bid_one.denomination == self.responder_bid_one.denomination:
            bid = self._at_three_overcall_support()
        elif self._is_balanced_or_no_fit_with_responder():
            bid = self.nt_bid(3, '0296')
        else:
            bid = self._at_three_overcall_no_support()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _responder_at_three_nt_overcall(self):
        """Return bid after responder has jumped to 3NT after an overcall."""
        if self.hcp >= 18 and self.nt_level <= 4:
            bid = self.nt_bid(4, '0297')
        elif self._has_six_card_major_and_can_bid_game():
            bid = self.suit_bid(4, self.longest_suit, '0298')
        else:
            bid = Pass('0299')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _at_three_overcall_support(self):
        """Return bid with 3 level support after an overcall."""
        if self.hcp >= 14:
            bid = self._overcall_support_strong()
        else:
            bid = Pass('0300')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _overcall_support_strong(self):
        """Return bid with 3 level support after overcall with strong hand."""
        hvp = self.support_points(self.bid_one.denomination)
        if self.bid_one.denomination.is_minor and hvp >= 18:
            bid = self._at_three_overcall_support_minor()
        elif self._is_balanced_minor_suit_can_bid_nt_game():
            bid = self.nt_bid(3, '0301')
        elif self.next_level(self.bid_one.denomination) <= 4:
            bid = self.suit_bid(4, self.bid_one.denomination, '0302')
        else:
            bid = Pass('0303')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _at_three_overcall_support_minor(self):
        """Return bid with 3 level support after overcall with strong minor."""
        if self._is_balanced_can_bid_nt_game():
            bid = self.nt_bid(3, '0304')
        elif self.next_level(self.bid_one.denomination) <= 5:
            bid = self.suit_bid(5, self.bid_one.denomination, '0305')
        else:


            bid = Pass('0911')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _at_three_overcall_no_support(self):
        """Return bid with no support, responder at 3 after overcall."""
        if self._has_six_card_major_and_can_bid_at_three_level():
            bid = self.suit_bid(3, self.bid_one.denomination, '0306')
        else:
            bid = self._overcall_no_support_minor()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _overcall_no_support_minor(self):
        """Return bid with no support, minor suit, after overcall."""
        responders_suit = self.responder_bid_one.denomination
        next_level = self.next_level(responders_suit)
        level = '4'
        if responders_suit.is_minor and self.hcp >= 15:
            level = '5'
        if (self.five_five and
                self.second_suit not in self.opponents_suits):
            bid = self.next_level_bid(self.second_suit, '0307')
        elif int(level) >= next_level:
            bid = self.suit_bid(level, responders_suit, '0308')
        else:
            bid = Pass('0309')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _responder_has_doubled(self):
        """Return bid after responder has doubled."""
        if (self.suit_length(self.second_suit) >= 4 and
                self.second_suit not in self.opponents_suits):
            bid = self.next_level_bid(self.second_suit, '0000')
        else:
            bid = self.next_level_bid(self.longest_suit, '0000')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    # Various utility functions

    def _get_second_suit_for_opener(self):
        """Return suit to bid after overcall."""
        if self.shape == [5, 4, 4, 0]:
            if self.second_suit not in self.opponents_suits:
                suit = self.second_suit
            elif self.third_suit not in self.opponents_suits:
                suit = self.third_suit
            else:
                suit = self.longest_suit
        elif self.second_suit in self.opponents_suits:
            suit = self.longest_suit
        else:
            suit = self.second_suit
        return suit

    def _get_level_with_support(self, suit):
        """Return appropriate level with partner support."""
        # use distribution points
        hand_value_points = (self.hcp +
                             self.support_shape_points(suit))
        point_list = [13, 17, 19]
        if suit.is_major:
            game = 4
        else:
            game = 5
        # quantitative_raise with point_list = [13, 16, 19]
        # means single raise if points between 13 and 15
        # jump  with points between 16 and 18 etc.
        level = self.quantitative_raise(hand_value_points,
                                        self.responder_bid_one.level,
                                        point_list, game)
        return level

    def _get_suit_support_for_responder(self):
        """Return suit if opener can support responder."""
        suit_to_bid = self.responder_bid_one.denomination
        if self.responder_bid_one.is_double:
            if self.shape[0] >= 6:
                suit_to_bid = self.longest_suit
            else:
                suit_to_bid = self.second_suit
        return suit_to_bid

    def _get_suit_no_support_five_five(self):
        """Return the suit to bid if no support and hand is 5/5."""
        suit_to_bid = self._other_five_card_suit()
        barrier_is_broken = self.barrier_is_broken(self.bid_one,
                                                   self.next_level_bid(suit_to_bid))

        can_break_barrier = (self.hcp >= 15 or
                             (self.hcp >= 14 and self.responder_bid_one.is_nt))
        if barrier_is_broken and not can_break_barrier:
            suit_to_bid = self.bid_one.denomination
        return suit_to_bid

    def _get_best_suits(self):
        """Return the *ordered* first and second suits."""
        first_suit = self.longest_suit
        second_suit = self.second_suit
        if self.five_five:
            first_suit = self.bid_one.denomination
            second_suit = self._other_five_card_suit()
        return first_suit, second_suit

    def _select_suit_if_six_four(self):
        """Return selected suit with 6/4 shape."""
        suit = self.longest_suit
        if self.shape[1] >= 4:
            if self.second_suit not in self.opponents_suits:
                second_suit = self.second_suit
                suit = self.cheaper_suit(suit, second_suit)
        return suit

    def _other_five_card_suit(self):
        """Return unbid 5 card suit in 5/5 hands."""
        suit_one = self.bid_one.denomination
        if suit_one == self.longest_suit:
            suit = self.second_suit
            if suit in self.opponents_suits:
                suit = self.longest_suit
        else:
            suit = self.longest_suit
            if suit in self.opponents_suits:
                suit = self.second_suit
        return suit

    def _raise_level_with_five_four_strong(self, suit_to_bid, test_bid):
        """Return the raise level with a strong 5/4 hand."""
        if self.hcp >= 19:
            raise_level = 1
        elif self.hcp < 16:
            raise_level = 0
        elif test_bid.level > 3:
            raise_level = 0
        elif suit_to_bid == self.bid_one.denomination:
            raise_level = 0
        elif suit_to_bid.rank > self.bid_one.denomination.rank:
            raise_level = 0
        else:
            raise_level = 1
        return raise_level

    def _five_four_weak_suit(self, suit, cannot_show_second_suit):
        """Return selected suit for weak 5/4 hands."""
        if suit < self.bid_one.denomination or not cannot_show_second_suit:
            suit_to_bid = suit
        elif self.bid_one.denomination == self.club_suit and suit == self.spade_suit:
            suit_to_bid = self.spade_suit
        else:
            suit_to_bid = self.bid_one.denomination
        return suit_to_bid

    def _get_suit_with_five_four(self):
        """Return suit with 5/4 hands."""
        if self.bid_one.denomination == self.longest_suit:
            suit = self.second_suit
            if self.shape[1] == 4 and self.shape[2] == 4:
                suit = self.cheaper_suit(self.ordered_holding[1][1],
                                         self.ordered_holding[2][1])
            if not self.can_bid_suit(suit):
                suit = self.longest_suit
        else:
            suit = self.longest_suit
            if suit in self.opponents_suits:
                suit = self.second_suit
        return suit

    def _cheapest_four_card_suit(self):
        """Return the cheapest 4 card suit."""
        suits = []
        for suit in self.suits:
            if self.suit_length(suit) == 4:
                suits.append(suit)
        cheapest = None
        for suit in suits:
            if cheapest:
                if suit < cheapest:
                    cheapest = suit
            else:
                cheapest = suit
        return cheapest

    def _balanced_bid_level(self):
        """Return bid level for balanced rebids."""
        bid_level = 0
        if self.nt_level == 1:
            if self.hcp >= 19:
                bid_level = 3
            elif self.hcp >= 17:
                bid_level = 2
            else:
                bid_level = 1
        elif self.nt_level == 2:
            if self._has_seventeen_points_and_no_support_from_responder():
                bid_level = 3
            elif self.hcp >= 18 and self.responder_bid_one.level == 2:
                bid_level = 3
            else:
                bid_level = 2
        if bid_level == 0 and self._balanced_jump():
            bid_level = 3
        return bid_level

    def _select_suit_for_four_four_four_one(self):
        """Select suit with 4441 hands after partner bids singleton suit."""
        singleton_suit = self.suit_shape[3]
        rank = singleton_suit.rank
        rank = (rank+1) % 4
        if self.suits[rank] in self.opponents_suits:
            rank = (rank+1) % 4
            if rank == self.bid_one.denomination.rank:
                rank = (rank+1) % 4
        suit = self.suits[rank]
        return suit

    # Various boolean tests

    def _balanced_jump(self):
        """Return True if there has been a jump bid."""
        if self.overcaller_bid_one.is_value_call:
            result = self.is_jump(self.overcaller_bid_one, self.responder_bid_one)
        else:
            result = self.is_jump(self.bid_one, self.responder_bid_one)
        return result

    def _weak_and_able_to_repeat_suit(self, cannot_show_second_suit, responder_jumped):
        """Return True if weak and able to repeat suit"""
        result = (cannot_show_second_suit and
                  self.hcp <= 14 and
                  not responder_jumped and
                  (not self.overcall_made or
                   self.shape[0] >= 6))
        return result

    def _has_strong_minor_responder_has_jumped(self):
        """Return True with strong hand after partner jumps and no major."""
        result = (self.hcp >= 16 and
                  self.is_jump(self.bid_one, self.responder_bid_one) and
                  self.bid_one.denomination.is_minor and
                  self.responder_bid_one.denomination.is_minor and
                  not self.overcaller_bid_one.is_value_call)
        return result

    def _competitive_auction_weak_with_minor(self):
        """Return True if competitive auction, no support and minor."""
        result = ((self.advancer_bid_one.is_value_call or
                   self.overcaller_bid_one.is_nt) and
                  self.bid_one.denomination.is_minor and
                  self.hcp <= 13 and
                  self.shape[0] <= 5)
        return result

    def _competitive_auction_with_support_for_responder(self):
        """Return True if competitive auction and support for partner's major."""
        result = ((self.overcaller_bid_one.is_value_call or
                   self.advancer_bid_one.is_value_call) and
                  self.responder_bid_one.denomination.is_major and
                  self.responder_bid_one.level >= 2 and
                  self.suit_length(self.responder_bid_one.denomination) >= 3)
        return result

    def _responder_has_bid_nt_game_over_minor(self):
        """Return True of responder bid 3NT over a minor."""
        result = (self.responder_bid_one.name == '3NT' and
                  self.bid_one.denomination.is_minor)
        return result

    def _is_shapely_intermediate_hand(self):
        """Return True if two suits and 14+ points."""
        result = (self.shape[0] + self.shape[1] >= 9 and
                  (self.responder_bid_one.name != '1NT' or
                   self.hcp >= 14))
        return result

    def _repeatable_five_card_suit(self):
        """Return True if weak with repeatable five card suit."""
        first_suit = self.longest_suit
        level = self.next_level(first_suit)
        result = (self.hcp >= 16 and
                  self.shape[0] >= 5 and level <= 2 and
                  self.suit_points(first_suit) >= 5)
        return result

    def _weak_five_four(self):
        """Return True if weak but biddable 5/4."""
        second_suit = self.second_suit
        result = (self.hcp >= 12 and
                  self.five_four and
                  second_suit not in self.opponents_suits)
        return result

    def _three_card_support_for_responder_overcall_made(self):
        """Return True with support for responder and overcall made."""
        result = (self.overcall_made and
                  not self.overcaller_bid_one.is_double and
                  self.hcp == 14 and
                  self.suit_length(self.responder_bid_one.denomination) >= 3)
        return result

    def _four_card_support_for_responder_overcall_made(self):
        """Return True with support for responder and overcall made."""
        result = (self.hcp == 13 and
                  self.suit_length(self.responder_bid_one.denomination) >= 4 and
                  self.overcall_made and
                  not self.overcaller_bid_one.is_double)
        return result

    def _strong_with_second_suit(self, level):
        """Return True if strong and has biddable second suit."""
        second_suit = self.second_suit
        result = (self.hcp >= 16 and
                  self.five_four and
                  (self.shape[0] <= 5 or
                   self.suit_points(second_suit) >= 6) and
                  second_suit != self.overcaller_bid_one.denomination and
                  second_suit != self.advancer_bid_one.denomination and
                  self.overcaller_bid_one.denomination == self.advancer_bid_one.denomination and
                  level <= 3)
        return result

    def _is_strong_with_five_five(self, second_suit):
        """"Return True with 16 points and two five card suits."""
        result = (self.hcp >= 16 and
                  self.five_five_or_better and
                  second_suit not in self.opponents_suits)
        return result

    def _is_strong_with_five_four(self, second_suit):
        """"Return True with 16 points and five/four."""
        opening_next_level = self.next_level(self.my_last_bid.denomination)
        result = (self.hcp >= 16 and
                  self.five_four and
                  self.next_level(second_suit) <= opening_next_level and
                  self.next_level(second_suit) <= 3 and
                  second_suit not in self.opponents_suits)
        return result

    def _six_card_suit_weak(self):
        """Return True if weak with six card suit."""
        first_suit = self.longest_suit
        level = self.next_level(first_suit)
        result = (self.shape[0] >= 6 and
                  12 <= self.hcp <= 15 and
                  level <= 3 and
                  (level <= 2 or
                   self.suit_points(first_suit) >= 5) and
                  Bid(self.bid_history[-1]).is_pass)
        return result

    def _strong_and_balanced(self):
        """Return True if strong, balanced and got stoppers."""
        result = (self.hcp >= 17 and
                  self.is_balanced and
                  self.stoppers_in_bid_suits and
                  not self.overcaller_bid_one.is_nt and
                  (self.responder_bid_one.is_value_call or
                   self.nt_level <= 1))
        return result

    def _weak_six_cards(self):
        """Return True if weak with six card suit."""
        result = (self.hcp >= 12 and
                  self.shape[0] >= 6 and
                  self.suit_points(self.longest_suit) >= 5 and
                  self.longest_suit not in self.opponents_suits and
                  self.next_level(self.longest_suit) <= 3)
        return result

    def _can_rebid_five_card_suit_at_two_level(self):
        """Return True if 5 card suit can be rebid."""
        result = (self.shape[0] == 5 and
                  self.hcp >= 12 and
                  self.next_level(self.longest_suit) <= 2 and
                  not self.bid_one.is_nt and
                  not self.overcaller_bid_one.is_nt)
        return result

    def _can_rebid_seven_card_suit_at_three_level(self):
        """Return True if 7 card suit can be rebid."""
        result = (self.shape[0] >= 7 and
                  self.next_level(self.longest_suit) <= 3)
        return result

    def _overcaller_doubled_and_advancer_bid(self):
        """Return True if overcaller has doubled and advancer has bid."""
        result = (self.overcaller_bid_one.is_double and
                  not self.advancer_bid_one.is_pass and
                  self.hcp <= 15)
        return result

    def _can_rebid_spades(self, level):
        """Return True if opener can rebid spades."""
        result = False
        if self.responder_bid_one.name == '3S' and self.spades >= 3:
            result = True
        elif (self.spades >= 2 and
              self.advancer_bid_one.is_value_call and
              self.spade_suit not in self.opponents_suits and
              level <= 4):
            result = True
        if self.next_level(self.spade_suit) > 4:
            return False
        return result

    def _responder_shows_support_or_can_support_responder(self):
        """Return True if responder has shown support."""
        result = (self.responder_bid_one.denomination == self.bid_one.denomination or
                  self.suit_length(self.responder_bid_one.denomination) >= 4 or
                  (self.responder_bid_one.level == 2 and
                   self.responder_bid_one.is_major and
                   self.suit_length(self.responder_bid_one.denomination) >= 3))
        return result

    def _can_support_responders_major(self):
        """Return True can support responder's major."""
        result = (self.responder_bid_one.level >= 2 and
                  self.suit_length(self.responder_bid_one.denomination) >= 3 and
                  self.responder_bid_one.denomination.is_major and
                  self.overcall_made)
        return result

    def _has_six_card_minor_and_eighteen_points(self):
        """Return True with 18 points and six card suit."""
        result = (self.hcp >= 18 and
                  self.responder_bid_one.denomination.is_minor and
                  self.shape[0] >= 6 and
                  self.stoppers_in_bid_suits and
                  self.stoppers_in_unbid_suits and
                  self.nt_level <= 3)
        return result

    def _is_unbalanced_with_sixteen_points(self):
        """Return True if 16+ points and unbalanced."""
        result = (self.hcp >= 16 and
                  not self.is_balanced and
                  self .second_suit not in self.opponents_suits)
        return result

    def _can_support_nineteen_points(self):
        """Return True if 19+ points and support for responder."""
        result = (self.hcp >= 19 and
                  self.responder_bid_one.denomination == self.second_suit and
                  self.game_level(self.second_suit) >= self.next_level(self.second_suit))
        return result

    def _can_support_seventeen_points_and_void(self):
        """Return True if 17+ points and support for responder and a void."""
        result = (self.hcp >= 17 and
                  self.responder_bid_one.denomination == self.second_suit and
                  self.shape[3] == 0 and
                  self.game_level(self.second_suit) >= self.next_level(self.second_suit))
        return result

    def _suit_is_major_and_eighteen_points(self, suit_to_bid):
        """Return True if 18+ points and level <= 3."""
        result = (self.next_level(suit_to_bid) <= 3 and
                  suit_to_bid.is_major and
                  self.hcp >= 18)
        return result

    def _responder_support_with_jump(self):
        """Return True if responder supports and has jumped."""
        result = (self.responder_bid_one.denomination == self.bid_one.denomination and
                  self.is_jump(self.bid_one, self.responder_bid_one) and
                  self.nt_level <= 3)
        return result

    def _can_repeat_suit_and_seventeen_points(self, suit_to_bid):
        """Return True if can repeat opening suit and 17+ points."""
        result = (suit_to_bid == self.bid_one.denomination and
                  self.hcp >= 17 and
                  self.stoppers_in_bid_suits)
        return result

    def _has_five_four_and_sixteen_points(self, suit_to_bid):
        """Return True if 5/4 and 16+ points."""
        result = (suit_to_bid == self.bid_one.denomination and
                  self.hcp >= 16 and
                  self.five_four_or_better)
        return result

    def _five_four_strong_test_nt(self, test_bid):
        """Return True if NT bid is allowed."""
        responder_jumped = self.jump_bid_made(self.responder_bid_one)
        result = (test_bid.level == 3 and
                  self.nt_level <= 3 and
                  self.shape[0] < 6 and
                  self.is_balanced and
                  self.stoppers_in_bid_suits and
                  not responder_jumped)
        return result

    def _weak_five_four_with_weak_suit(self, suit_to_bid):
        """Return True if suit too weak to rebid."""
        value_one = (suit_to_bid == self.bid_one.denomination and
                     self.overcall_made and
                     not self.responder_bid_one.is_value_call and
                     self.suit_points(suit_to_bid) <= 4)
        value_two = (suit_to_bid == self.bid_one.denomination and
                     (self.overcaller_bid_one.is_nt or
                      self.advancer_bid_one.is_nt) and
                     self.suit_points(suit_to_bid) <= 4)
        result = (value_one or value_two and
                  (self.responder_bid_one.is_pass or
                   self.bid_history[-2] != 'P'))
        return result

    def _can_bid_at_appropriate_level(self, suit_to_bid, level):
        """Return True if possible to bid at appropriate level."""
        result = (level <= 2 or
                  ((self.hcp >= 13 or
                   suit_to_bid == self.bid_one.denomination) and
                   self.responder_bid_one.level == 2))
        return result

    def _weak_five_four_consider_bid(self, level):
        """Return True if a sound bid is possible for weak 5/4."""
        result = (level <= 2 or self.shape[0] >= 6 or
                  (self.responder_bid_one.is_nt and level <= 3) or
                  not self.overcall_made or
                  (self.jump_bid_made(self.responder_bid_one) and
                   self.responder_bid_one.denomination != self.bid_one.denomination))
        return result

    def _weak_after_one_nt_and_barrier_broken(self, cannot_show_second_suit):
        """Return True if barrier broken after responder bids 1NT."""
        result = (self.hcp <= 15 and
                  cannot_show_second_suit and
                  self.responder_bid_one.name == '1NT')
        return result

    def _six_four_responder_at_level_two(self):
        """"Return True if 6/4 and responder at level 2."""
        result = (self.six_four and
                  self.responder_bid_one.level == 2 and
                  self.levels_to_bid(self.bid_one,
                                     self.next_level_bid(self.second_suit)) <= 2 and
                  self.second_suit not in self.opponents_suits)
        return result

    def _strong_or_can_bid_at_level_two(self, suit_to_bid):
        """Return True if can bid at level 2 or strong."""
        result = (self.next_level(suit_to_bid) <= 2 or
                  self.hcp >= 16 or
                  (self.overcall_made and
                   self.suit_length(self.overcaller_bid_one.denomination) == 0))
        return result

    def _strong_five_four_or_better(self):
        """Return True if strong and 5/4 or better."""
        result = (self.hcp >= 15 and
                  self.five_four_or_better and
                  self.second_suit not in self.opponents_suits)
        return result

    def _is_strong_with_seven_card_major(self):
        """Return True if strong with seven card major."""
        result = (self.hcp >= 16 and
                  self.shape[0] >= 7 and
                  self.longest_suit.is_major and
                  self.next_level(self.longest_suit) <= self.longest_suit.game_level)
        return result

    def _is_strong_six_four_with_major(self):
        """Return True if strong with six card major."""
        result = (self.hcp >= 16 and
                  self.second_suit.is_major and
                  self.second_suit not in self.opponents_suits and
                  self.shape[1] >= 4)
        return result

    def _weak_with_overcall_or_next_level_three(self):
        """Return True if overcall or next level greater than 2."""
        result = ((self.overcall_made and
                   self.hcp <= 15) or
                  self.next_level(self.longest_suit) > 2)
        return result

    def _has_minor_can_support_responders_major(self):
        """Return True if suit is a minor and can support responder's major."""
        result = (self.longest_suit.is_minor and
                  self.responder_bid_one.denomination.is_major and
                  self.suit_length(self.responder_bid_one.denomination) >= 4)
        return result

    def _has_six_card_major_and_responder_bid_two_nt(self):
        """Return True if six card major after responder bids NT."""
        result = (self.responder_bid_one.name == '2NT' and
                  self.shape[0] >= 6 and
                  self.longest_suit.is_major)
        return result

    def _is_strong_with_solid_seven_card_suit(self):
        """Return True if strong with solid 7 card suit."""
        result = (self.hcp >= 15 and
                  self.shape[0] >= 7 and
                  self.solid_suit_honours(self.longest_suit))
        return result

    def _has_five_card_suit_fifteen_points_level_two(self):
        """Return True if intermediate with 5 card suit."""
        result = (self.hcp >= 15 and
                  self.shape[0] >= 5 and
                  self.next_level(self.longest_suit) <= 2)
        return result

    def _has_seven_card_suit_and_weak(self):
        """Return True if weak with 7 card suit."""
        result = (self.hcp >= 11 and
                  self.shape[0] >= 7 and
                  self.next_level(self.longest_suit) <= 3)
        return result

    def _has_six_card_suit_responder_at_level_two(self):
        """Return True if 6 card suit and responder at level 2."""
        result = (self.shape[0] >= 6 and
                  self.responder_bid_one.level >= 2 and
                  self.next_level(self.longest_suit) <= 3)
        return result

    def _has_fifteen_points_and_level_is_one(self, level):
        """Return True if 15+ points and level is 1."""
        result = self.hcp >= 15 and level >= self.nt_level
        return result

    def _barrier_would_break_and_fewer_that_sixteen(self, suit, test_bid):
        """Return True if barrier broken and fewer than 16 points."""
        cannot_show_second_suit = (self.barrier_is_broken(self.bid_one, test_bid) and
                                   not self.is_jump(self.bid_one, self.responder_bid_one))
        result = ((self.hcp <= 16 and cannot_show_second_suit) or suit in self.opponents_suits)
        return result

    def _strong_can_bid_nt(self, bid_level):
        """Return True if it appropriate to bid NT."""
        result = (0 <= bid_level <= 3 and
                  bid_level >= self.nt_level and
                  (not self._weak_partner() or self.hcp >= 18))
        return result

    def _no_fit_fifteen_points_responder_level_two(self):
        """Return True with 15+points and no fit and responder at level 2."""
        result = (self.responder_bid_one.denomination != self.bid_one.denomination and
                  self.responder_bid_one.level >= 2 and
                  self.hcp >= 15 and
                  self.stoppers_in_bid_suits and
                  self.nt_level <= 3)
        return result

    def _has_seventeen_points_and_no_support_from_responder(self):
        """Return True with 17+ points and no support."""
        result = (self.hcp >= 17 and
                  self.bid_one.denomination != self.responder_bid_one.denomination and
                  not self.overcall_made)
        return result

    def _is_balanced_or_no_fit_with_responder(self):
        """Return True if balanced or no fit with responder."""
        result = ((self.is_balanced or
                   self.suit_length(self.responder_bid_one.denomination) <= 2) and
                  self.stoppers_in_bid_suits and
                  self.nt_level <= 3)
        return result

    def _weak_partner(self):
        """Return True if partner has indicated a weak hand."""
        result = (self.responder_bid_one.level == 1 and
                  (self.overcaller_bid_one.level == 2 or
                   self.overcaller_bid_one.is_double or
                   self.advancer_bid_one.is_nt or
                   self.advancer_bid_one.level == 2) or
                  (self._responder_bids_openers_suit_at_lowest_level() and
                   self.hcp < 17))
        return result

    def _responder_bids_openers_suit_at_lowest_level(self):
        """Return True if responder bids opener's suit at lowest level."""
        result = (self.my_last_bid.denomination == self.responder_bid_one.denomination and
                  self.my_last_bid.level+1 == self.responder_bid_one.level)
        return result

    def _has_six_card_major_and_can_bid_game(self):
        """Return True with 6 card major and below game level."""
        result = (self.shape[0] >= 6 and
                  self.longest_suit.is_major and
                  self.next_level(self.longest_suit) <= 4)
        return result

    def _has_six_card_major_and_can_bid_at_three_level(self):
        """Return True with 6 card major and can_bid_at 3 level."""
        result = (self.shape[0] >= 6 and
                  self.longest_suit.is_major and
                  self.next_level(self.bid_one.denomination) <= 3)
        return result

    def _is_balanced_minor_suit_can_bid_nt_game(self):
        """Return True if suit is manor, balanced and can bid 3NT."""
        result = (self.bid_one.denomination.is_minor and
                  self._is_balanced_can_bid_nt_game())
        return result

    def _is_balanced_can_bid_nt_game(self):
        result = (self.is_balanced and
                  self.stoppers_in_bid_suits and
                  self.nt_level <= 3)
        return result

    def _can_bid_two_nt(self):
        """Return True if can bid 2NT."""
        result = ((self.hcp >= 17 or self.nt_level == 2) and
                  self.nt_level <= 2 and
                  self.stoppers_in_bid_suits)
        return result

    def _powerful_and_has_four_cards_in_unbid_major(self, unbid_major):
        """Return True if powerful hand with 4 card unbid major."""
        result = (unbid_major is not None and
                  unbid_major not in self.opponents_suits and
                  (self.five_four_or_better or
                   self.my_last_bid.is_minor))
        return result

    def _weak_overcall_responder_level_one(self, ):
        """Return True if weak and responder has bid at level 1."""
        result = (self.hcp <= 12 and
                  self.overcall_made and
                  self.responder_bid_one.level == 1)
        return result

    def _fffo_strong_four_card_support(self):
        """Return True if strong and can support partner suit."""
        result = (self.responder_bid_one.level == 2 and
                  self.shape == [4, 4, 4, 1] and
                  self.suit_length(self.responder_bid_one.denomination) >= 4 and
                  self.hcp >= 17)
        return result

    def _can_bid_three_nt(self):
        """Return True if can support NTs."""
        result = (self.responder_bid_one.is_nt and
                  self.shape[0] >= 5 and
                  self.nt_level <= 3 and
                  self.stoppers_in_bid_suits and
                  self.hcp >= 12)
        return result

    def _no_biddable_second_suit_and_support_for_partner(self):
        """Return True if Support for responder and second suit not biddable major."""
        suit = self.responder_bid_one.denomination
        result = (self.responder_bid_one.level >= 2 and
                  self.hcp >= 16 and
                  self.suit_length(suit) >= 4 and
                  self.next_level(suit) <= suit.game_level and
                  not (self.second_suit.is_major and self.shape[1] >= 4))
        return result

    def _can_bid_second_suit_after_nt(self):
        """Return True if can bid second suit after a 2NT response."""
        result = (self.five_four and
                  self.partner_bid_one.name == '2NT' and
                  self.second_suit.rank < self.longest_suit.rank and
                  self.second_suit not in self.opponents_suits)
        return result

    def _has_six_card_major_and_responder_bid_one_nt(self):
        """Return True if xxx."""
        result = (self.hcp >= 15 and
                  self.shape[0] >= 6 and
                  self.longest_suit.is_major and
                  self.responder_bid_one.name == '1NT')
        return result

    # def xxx(self):
    #     """Return True if xxx."""
    #     result = False
    #     return result

    # def xxx(self):
    #     """Return True if xxx."""
    #     result = False
    #     return result



"""Expose the classes in the API."""
from ._version import __version__

VERSION = __version__

from .source.hand import Hand
from .source.comments import comments, strategies, comment_xrefs, convert_text_to_html
from .source.strategy_xref import StrategyXref, strategy_descriptions
from .source.bidding import Bid, Pass, Double
from .source.player import Player



""" Bid for Game
    Acol OverCallerRespondersLaterBid module
"""

import inspect
from .bidding import Bid, Pass
from .hand import Hand


class AdvancersLaterBid(Hand):
    """BfG AdvancersRebid class."""
    def __init__(self, hand_cards, board):
        super().__init__(hand_cards, board)
        if len(self.bid_history) >= 11:
            self.overcaller_bid_one = Bid(self.bid_history[-10], '')
            self.overcaller_bid_two = Bid(self.bid_history[-6], '')
            self.overcaller_bid_three = Bid(self.bid_history[-2], '')
        else:
            self.overcaller_bid_one = Bid(self.bid_history[-6], '')
            self.overcaller_bid_two = None
        self.advancer_bid_one = Bid(self.bid_history[-8], '')
        self.trace = 0

    def suggested_bid(self):
        """Direct control to relevant method and return a Bid object."""
        if self._overcaller_shows_two_suits_after_double():
            bid = self.advancer_preference('0894')
            if (self.suit_length(bid.denomination) <= 2 and
                    self.nt_level <= 3):
                bid = self.nt_bid(3, '0239')
        elif self._has_two_card_support_for_overcallers_six_card_suit():
            if self.overcaller_bid_one.is_double:
                overcaller_bid = self.overcaller_bid_two
            else:
                overcaller_bid = self.overcaller_bid_one
            bid = self.bid_to_game(overcaller_bid.denomination, '0861')
        elif self._can_show_second_five_card_suit():
            bid = self. next_level_bid(self.second_suit, '0862')
        elif self._has_biddable_seven_card_suit():
            bid = self.next_level_bid(self.longest_suit, '0869')
        elif self.overcaller_bid_one.name == '1NT':
            bid = self._responses_to_overcall_of_one_nt()
        else:
            bid = Pass('0863')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _responses_to_overcall_of_one_nt(self):
        """Handle responses after overcaller opens 1NT."""
# TODO: this is just a stub - much more logic required
        if self.advancer_bid_one.name == '2C':
            bid = self._responses_after_stayman()
        else:
            bid = Pass('0000')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _responses_after_stayman(self):
        """Handle responses after Stayman."""
# TODO: this is just a stub - much more logic required
        if (self.overcaller_bid_two.denomination == self.heart_suit and
                self.overcaller_bid_three.denomination == self.spade_suit):
            if self.spades >= 4:
                bid = self.spade_bid(4, '0240')
            elif self.nt_level <= 3:
                bid = self.nt_bid(3, '0000')
            else:
                bid = Pass('0000')
        else:
            bid = Pass('0000')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    # Various utility functions

    # Various boolean functions

    def _overcaller_shows_two_suits_after_double(self):
        """Return True if overcaller shows two suits after double."""
        result = (self.overcaller_bid_one.is_double and
                  self.overcaller_bid_two.is_suit_call and
                  self.overcaller_bid_three.is_suit_call and
                  self.overcaller_bid_two.denomination != self.overcaller_bid_three.denomination and
                  not self.overcaller_bid_three.is_game)
        return result

    def _has_two_card_support_for_overcallers_six_card_suit(self):
        """Return True if 2 card support for overcallers six card suit."""
        if self.opponents_at_game:
            return False
        if self.overcaller_bid_one.is_double:
            overcaller_bid = self.overcaller_bid_two
        else:
            overcaller_bid = self.overcaller_bid_one
        if self.overcaller_bid_two.is_double:
            overcaller_bid = self.overcaller_bid_three
        else:
            overcaller_bid = self.overcaller_bid_two
        if overcaller_bid.is_pass or overcaller_bid.is_double:
            return False
        game_level = self.bid_to_game(overcaller_bid.denomination).level
        result = (overcaller_bid == self.overcaller_bid_three.denomination and
                  self.suit_length(self.overcaller_bid_one.suit) >= 2 and
                  self.hcp >= 9 and
                  not self.overcaller_bid_three.is_game and
                  self.next_level(overcaller_bid.denomination) <= game_level)
        return result

    def _can_show_second_five_card_suit(self):
        """Return True if can show second 5 card suit."""
        result = (self.five_five and
                  not self.advancer_bid_one.is_pass and
                  self.second_suit not in self.opponents_suits and
                  self.next_level(self.second_suit) <= 4 and
                  (not self.overcaller_bid_three.is_game or
                   self.second_suit.is_major))
        return result

    def _has_biddable_seven_card_suit(self):
        """Return True if can show 7 card suit."""
        result = (not self.overcaller_bid_three.is_pass and
                  self.shape[0] >= 7 and
                  self.next_level(self.longest_suit) <= 4 and
                  self.longest_suit not in self.opponents_suits)
        return result

    # def xxx(self):
    #     """Return True if xxx."""
    #     result = False
    #     return result
    #
    # def xxx(self):
    #     """Return True if xxx."""
    #     result = False
    #     return result
